from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.port import Port
from glayout.pdk.mappedpdk import MappedPDK
from typing import Optional
from glayout.primitives.via_gen import via_stack, via_array
from gdsfactory.components.rectangle import rectangle
from glayout.pdk.util.comp_utils import evaluate_bbox, align_comp_to_port
from glayout.pdk.util.port_utils import assert_port_manhattan, set_port_orientation, add_ports_perimeter
from gdstk import rectangle as primitive_rectangle


@cell
def straight_route(
	pdk: MappedPDK,
	edge1: Port,
	edge2: Port,
	glayer1: Optional[str] = None,
	width: Optional[float] = None,
	glayer2: Optional[str] = None,
	via1_alignment: Optional[tuple[str, str]] = None,
	via1_alignment_layer: Optional[str] = None,
	via2_alignment: Optional[tuple[str, str]] = None,
	via2_alignment_layer: Optional[str] = None,
	fullbottom: Optional[bool] = False
) -> Component:
	"""extends a route from edge1 until perpindicular with edge2, then places a via
	This depends on the orientation of edge1 and edge2
	if edge1 has the same orientation as edge2, the generator will rotate edge2 180 degrees
	Will not modify edge1 or edge2
	
	DOES NOT REQUIRE:
	edge2 is directly inline with edge1
	
	example:
	           edge2
	             
	edge1--------
	
	args:
	pdk to use
	edge1, edge2 Ports
	glayer1 = defaults to edge1.layer, layer of the route.
	****If not edge1.layer, a via will be placed
	glayer2 = defaults to edge2.layer, end layer of the via
	width = defaults to edge1.width
	via1_alignment = alignment of the via on edge1
	via2_alignment = alignment of the via on edge2
	****defaults to an orientation that is aligned to the orientation of the port.
	
	Ports:
	route_...all edges of the rectangle path
	"""
	#TODO: error checking
	width = width if width else edge1.width
	glayer1 = glayer1 if glayer1 else pdk.layer_to_glayer(edge1.layer)
	front_via = None
	if glayer1 != pdk.layer_to_glayer(edge1.layer):
		front_via = via_stack(pdk,glayer1,pdk.layer_to_glayer(edge1.layer),fullbottom=fullbottom)
	glayer2 = glayer2 if glayer2 else pdk.layer_to_glayer(edge2.layer)
	assert_port_manhattan([edge1,edge2])
	if edge1.orientation == edge2.orientation:
		edge2 = set_port_orientation(edge2,edge2.orientation,flip180=True)
	pdk.activate()
	# find extension length and direction
	edge1_is_EW = bool(round(edge1.orientation + 90) % 180)
	if edge1_is_EW:
		startx = edge1.center[0]
		endx = edge2.center[0]
		extension = endx-startx
		viaport_name = "route_E" if extension > 0 else "route_W"
		alignment = ("r","c") if extension > 0 else ("l","c")
		size = (abs(extension),width)
	else:
		starty = edge1.center[1]
		endy = edge2.center[1]
		extension = endy-starty
		viaport_name = "route_N" if extension > 0 else "route_S"
		alignment = ("c","t") if extension > 0 else ("c","b")
		size = (width,abs(extension))
	# create route and via
	route = Component()
	route.add_polygon(primitive_rectangle((0,0),size,*pdk.get_glayer(glayer1)))
	add_ports_perimeter(route,layer=pdk.get_glayer(glayer1),prefix="route_")
	out_via = via_stack(pdk,glayer1,glayer2,fullbottom=fullbottom) if glayer1 != glayer2 else None
	# place route and via
	straightroute = Component()
	for i, edge in enumerate([edge1,edge2]):
		temp = via1_alignment if i == 0 else via2_alignment
		if temp is None:
			if round(edge.orientation) == 0:# facing east
				temp = ("l", "c")
			elif round(edge.orientation) == 180:# facing west
				temp = ("r", "c")
			elif round(edge.orientation) == 270:# facing south
				temp = ("c", "t")
			elif round(edge.orientation) == 90:#facing north
				temp = ("c", "b")
			else:
				raise ValueError("port must be vertical or horizontal")
		via1_alignment = temp if i == 0 else via1_alignment
		via2_alignment = temp if i == 1 else via2_alignment
	route_ref = align_comp_to_port(route,edge1,alignment=alignment)
	straightroute.add_ports(route_ref.get_ports_list())
	straightroute.add(route_ref)
	if out_via is not None:
		alignlayer2 = pdk.get_glayer(glayer1) if via2_alignment_layer is None else pdk.get_glayer(via2_alignment_layer)
		straightroute.add(align_comp_to_port(out_via,route_ref.ports[viaport_name],layer=alignlayer2,alignment=via2_alignment))
	if front_via is not None:
		alignlayer1 = pdk.get_glayer(glayer1) if via1_alignment_layer is None else pdk.get_glayer(via1_alignment_layer)
		straightroute.add(align_comp_to_port(front_via,edge1,layer=alignlayer1,alignment=via1_alignment))
	return straightroute.flatten()





"""
usage: from mappedpdk import MappedPDK
"""

from gdsfactory.pdk import Pdk
from gdsfactory.typings import Component, PathType, Layer
from pydantic import validator, StrictStr, ValidationError
from typing import ClassVar, Optional, Any, Union, Literal, Iterable, TypedDict
from pathlib import Path
from decimal import Decimal, ROUND_UP
import tempfile
import subprocess
from decimal import Decimal
from pydantic import validate_arguments
import xml.etree.ElementTree as ET

class MappedPDK(Pdk):
    """Inherits everything from the pdk class but also requires mapping to glayers
    glayers are generic layers which can be returned with get_glayer(name: str)
    has_required_glayers(list[str]) is used to verify all required generic layers are
    present"""

    valid_glayers: ClassVar[tuple[str]] = (
        "dnwell",
        "pwell",
        "nwell",
        "p+s/d",
        "n+s/d",
        "active_diff",
        "active_tap",
        "poly",
        "mcon",
        "met1",
        "via1",
        "met2",
        "via2",
        "met3",
        "via3",
        "met4",
        "via4",
        "met5",
        "capmet",
    )

    models: dict = {
        "nfet": "",
        "pfet": "",
        "mimcap": ""
    }

    glayers: dict[StrictStr, Union[StrictStr, tuple[int,int]]]
    # friendly way to implement a graph
    grules: dict[StrictStr, dict[StrictStr, Optional[dict[StrictStr, Any]]]]
    klayout_lydrc_file: Optional[Path] = None

    @validator("models")
    def models_check(cls, models_obj: dict[StrictStr, StrictStr]):
        for model in models_obj.keys():
            if not model in ["nfet","pfet","mimcap"]:
                raise ValueError(f"specify nfet, pfet, or mimcap models only")
        return models_obj

    @validator("glayers")
    def glayers_check_keys(cls, glayers_obj: dict[StrictStr, Union[StrictStr, tuple[int,int]]]):
        """force people to pick glayers from a finite set of string layers that you define
        checks glayers to ensure valid keys,type. Glayers must be passed as dict[str,str]
        if someone tries to pass a glayers dict that has a bad key, throw an error"""
        for glayer, mapped_layer in glayers_obj.items():
            if (not isinstance(glayer, str)) or (not isinstance(mapped_layer, Union[str, tuple])):
                raise TypeError("glayers should be passed as dict[str, Union[StrictStr, tuple[int,int]]]")
            if glayer not in cls.valid_glayers:
                raise ValueError(
                    "glayers keys must be one of generic layers listed in class variable valid_glayers"
                )
        return glayers_obj

    @validator("klayout_lydrc_file")
    def lydrc_file_exists(cls, lydrc_file_path):
        """Check that lydrc_file_path exists if not none"""
        if lydrc_file_path != None and not lydrc_file_path.is_file():
            raise ValueError(".lydrc script: the path given is not a file")
        return lydrc_file_path

    @validate_arguments
    def drc(
        self,
        layout: Component | PathType,
        output_dir_or_file: Optional[PathType] = None,
    ):
        """Returns true if the layout is DRC clean and false if not
        Also saves detailed results to output_dir_or_file location as lyrdb
        layout can be passed as a file path or gdsfactory component"""
        if not self.klayout_lydrc_file:
            raise NotImplementedError("no drc script for this pdk")
        # find layout gds file path
        tempdir = None
        if isinstance(layout, Component):
            tempdir = tempfile.TemporaryDirectory()
            layout_path = Path(layout.write_gds(gdsdir=tempdir.name)).resolve()
        elif isinstance(layout, PathType):
            layout_path = Path(layout).resolve()
        else:
            raise TypeError("layout should be a Component, Path, or string")
        if not layout_path.is_file():
            raise ValueError("layout must exist, the path given is not a file")
        # find report file path, if None then use current directory
        report_path = (
            Path(output_dir_or_file).resolve()
            if output_dir_or_file
            else Path.cwd().resolve()
        )
        if report_path.is_dir():
            report_path = Path(
                report_path
                / str(
                    self.name
                    + layout_path.name.replace(layout_path.suffix, "")
                    + "_drcreport.lyrdb"
                )
            )
        elif not report_path.is_file():
            raise ValueError("report_path must be file or dir")
        # run klayout drc
        drc_args = [
            "klayout",
            "-b",
            "-r",
            str(self.klayout_lydrc_file),
            "-rd",
            "input=" + str(layout_path),
            "-rd",
            "report=" + str(report_path),
        ]
        rtr_code = subprocess.Popen(drc_args).wait()
        if rtr_code:
            raise RuntimeError("error running klayout DRC")
        # clean up and return
        if tempdir:
            tempdir.cleanup()
        # there is a drc parsing open-source at:
        # https://github.com/google/globalfoundries-pdk-libs-gf180mcu_fd_pr/blob/main/rules/klayout/drc
        # eventually I can return more info on the drc run, but for now just void and view the lyrdb in klayout

        # Open DRC output XML file
        drc_tree = ET.parse(report_path.resolve())
        drc_root = drc_tree.getroot()
        if drc_root.tag != "report-database":
            raise TypeError("DRC report file is not a valid report-database")
        # Check if DRC passed
        drc_error_count = len(drc_root[7])
        return (drc_error_count == 0)

    @validate_arguments
    def has_required_glayers(self, layers_required: list[str]):
        """Raises ValueError if any of the generic layers in layers_required: list[str]
        are not mapped to anything in the pdk.glayers dictionary
        also checks that the values in the glayers dictionary map to real Pdk layers"""
        for layer in layers_required:
            if layer not in self.glayers:
                raise ValueError(
                    f"{layer!r} not in self.glayers {list(self.glayers.keys())}"
                )
            if isinstance(self.glayers[layer], str):
                self.validate_layers([self.glayers[layer]])
            elif not isinstance(self.glayers[layer], tuple):
                raise TypeError("glayer mapped value should be str or tuple[int,int]")


    @validate_arguments
    def layer_to_glayer(self, layer: tuple[int, int]) -> str:
        """if layer provided corresponds to a glayer, will return a glayer
        else will raise an exception
        takes layer as a tuple(int,int)"""
        # lambda for finding last matching key in dict from val
        find_last = lambda val, d: [x for x, y in d.items() if y == val].pop()
        if layer in self.glayers.values():
            return find_last(layer, self.glayers)
        elif self.layers is not None:
            # find glayer verfying presence along the way
            pdk_real_layers = self.layers.values()
            if layer in pdk_real_layers:
                layer_name = find_last(layer, self.layers)
                if layer_name in self.glayers.values():
                    glayer_name = find_last(layer_name, self.glayers)
                else:
                    raise ValueError("layer does not correspond to a glayer")
            else:
                raise ValueError("layer is not a layer present in the pdk")
            return glayer_name
        else:
            raise ValueError("layer might not be a layer present in the pdk")

    # TODO: implement LayerSpec type
    @validate_arguments
    def get_glayer(self, layer: str) -> Layer:
        """Returns the pdk layer from the generic layer name"""
        direct_mapping = self.glayers[layer]
        if isinstance(direct_mapping, tuple):
            return direct_mapping
        else:
            return self.get_layer(direct_mapping)

    @validate_arguments
    def get_grule(
        self, glayer1: str, glayer2: Optional[str] = None, return_decimal = False
    ) -> dict[StrictStr, Union[float,Decimal]]:
        """Returns a dictionary describing the relationship between two layers
        If one layer is specified, returns a dictionary with all intra layer rules"""
        if glayer1 not in MappedPDK.valid_glayers:
            raise ValueError("get_grule, " + str(glayer1) + " not valid glayer")
        # decide if two or one inputs and set rules_dict accordingly
        rules_dict = None
        if glayer2 is not None:
            if glayer2 not in MappedPDK.valid_glayers:
                raise ValueError("get_grule, " + str(glayer2) + " not valid glayer")
            rules_dict = self.grules.get(glayer1, dict()).get(glayer2)
            if rules_dict is None or rules_dict == {}:
                rules_dict = self.grules.get(glayer2, dict()).get(glayer1)
        else:
            glayer2 = glayer1
            rules_dict = self.grules.get(glayer1, dict()).get(glayer1)
        # error check, convert type, and return
        if rules_dict is None or rules_dict == {}:
            raise NotImplementedError(
                "no rules found between " + str(glayer1) + " and " + str(glayer2)
            )
        for rule in rules_dict:
            if type(rule) == float and return_decimal:
                rules_dict[rule] = Decimal(str(rule))
        return rules_dict

    @classmethod
    def is_routable_glayer(cls, glayer: StrictStr):
        return any(hint in glayer for hint in ["met", "active", "poly"])

    # TODO: implement
    @classmethod
    def from_gf_pdk(
        cls,
        gfpdk: Pdk,
        **kwargs
    ):
        """Construct a mapped pdk from an existing pdk and the extra parts of MappedPDK
        grid is the grid size in nm"""
        # input type and value validation
        if not isinstance(gfpdk, Pdk):
            raise TypeError("from_gf_pdk: gfpdk arg only accepts GDSFactory pdk type")
        # create argument dictionary
        passargs = dict()
        # pdk args
        passargs["name"]=gfpdk.name
        #passargs["cross_sections"]=gfpdk.cross_sections
        #passargs["cells"]=gfpdk.cells
        #passargs["symbols"]=gfpdk.symbols
        #passargs["default_symbol_factory"]=gfpdk.default_symbol_factory
        #passargs["containers"]=gfpdk.containers
        #passargs["base_pdk"]=gfpdk.base_pdk
        #passargs["default_decorator"]=gfpdk.default_decorator
        passargs["layers"]=gfpdk.layers
        #passargs["layer_stack"]=gfpdk.layer_stack
        #passargs["layer_views"]=gfpdk.layer_views#??? layer view broken???
#        passargs["layer_transitions"]=gfpdk.layer_transitions
#        passargs["sparameters_path"]=gfpdk.sparameters_path
#        passargs["modes_path"]=gfpdk.modes_path
#        passargs["interconnect_cml_path"]=gfpdk.interconnect_cml_path
#        passargs["warn_off_grid_ports"]=gfpdk.warn_off_grid_ports
#        passargs["constants"]=gfpdk.constants
#        passargs["materials_index"]=gfpdk.materials_index
#        passargs["routing_strategies"]=gfpdk.routing_strategies
#        passargs["circuit_yaml_parser"]=gfpdk.circuit_yaml_parser
#        passargs["gds_write_settings"]=gfpdk.gds_write_settings
#        passargs["oasis_settings"]=gfpdk.oasis_settings
#        passargs["cell_decorator_settings"]=gfpdk.cell_decorator_settings
#        passargs["bend_points_distance"]=gfpdk.bend_points_distance
        # MappedPDK args override existing args
        passargs.update(kwargs)
        # create and return MappedPDK
        mappedpdk = MappedPDK(**passargs)
        return mappedpdk

    # util methods
    @validate_arguments
    def util_max_metal_seperation(self, metal_levels: Union[list[int],list[str], str, int] = range(1,6)) -> float:
        """returns the maximum of the min_seperation rule for all layers specfied
        although the name of this function is util_max_metal_seperation, layers do not have to be metals
        you can specify non metals by using metal_levels=list of glayers
        if metal_levels is list of int, integers are converted to metal levels
        if a single int is provided, all metals below and including that int level are considerd
        by default this function returns the maximum metal seperation of metals1-5
        """
        if type(metal_levels)==int:
            metal_levels = range(1,metal_levels+1)
        metal_levels = metal_levels if isinstance(metal_levels,Iterable) else [metal_levels]
        if len(metal_levels)<1:
            raise ValueError("metal levels cannot be empty list")
        if type(metal_levels[0])==int:
            metal_levels = [f"met{i}" for i in metal_levels]
        sep_rules = list()
        for met in metal_levels:
            sep_rules.append(self.get_grule(met)["min_separation"])
        return self.snap_to_2xgrid(max(sep_rules))

    @validate_arguments
    def snap_to_2xgrid(self, dims: Union[list[Union[float,Decimal]], Union[float,Decimal]], return_type: Literal["decimal","float","same"]="float", snap4: bool=False) -> Union[list[Union[float,Decimal]], Union[float,Decimal]]:
        """snap all numbers in dims to double the grid size.
        This is useful when a generator accepts a size or dimension argument
        because there is a chance the cell may be centered (resulting in off grid components)
        args:
        dims = a list OR single number specifying the dimensions to snap to grid
        return_type = return a decimal, float, or the same type that was passed to the function
        snap4: snap to 4xgrid (Defualt false)
        """
        dims = dims if isinstance(dims, Iterable) else [dims]
        dimtype_in = type(dims[0])
        dims = [Decimal(str(dim)) for dim in dims] # process in decimals
        grid = 2 * Decimal(str(self.grid_size))
        grid = grid if grid else Decimal('0.001')
        grid = 2*grid if snap4 else grid
        # snap dims to grid
        snapped_dims = list()
        for dim in dims:
            snapped_dim = grid * (dim / grid).quantize(1, rounding=ROUND_UP)
            snapped_dims.append(snapped_dim)
        # convert to correct type
        if return_type=="float" or (return_type=="same" and dimtype_in==float):
            snapped_dims = [float(snapped_dim) for snapped_dim in snapped_dims]
        # correctly return list or single element
        return snapped_dims[0] if len(snapped_dims)==1 else snapped_dims




from gdsfactory.cell import cell, clear_cache
from gdsfactory.component import Component, copy
from gdsfactory.component_reference import ComponentReference
from gdsfactory.components.rectangle import rectangle
from glayout.pdk.mappedpdk import MappedPDK
from typing import Optional, Union
from glayout.primitives.fet import nmos, pmos, multiplier
from glayout.components.diff_pair import diff_pair
from glayout.primitives.guardring import tapring
from glayout.primitives.mimcap import mimcap_array, mimcap
from glayout.routing.L_route import L_route
from glayout.routing.c_route import c_route
from glayout.primitives.via_gen import via_stack, via_array
from gdsfactory.routing.route_quad import route_quad
from glayout.pdk.util.comp_utils import evaluate_bbox, prec_ref_center, movex, movey, to_decimal, to_float, move, align_comp_to_port, get_padding_points_cc
from glayout.pdk.util.port_utils import rename_ports_by_orientation, rename_ports_by_list, add_ports_perimeter, print_ports, set_port_orientation, rename_component_ports
from glayout.routing.straight_route import straight_route
from glayout.pdk.util.snap_to_grid import component_snap_to_grid
from pydantic import validate_arguments
from glayout.placement.two_transistor_interdigitized import two_nfet_interdigitized

from glayout.components.diff_pair_cmirrorbias import diff_pair_ibias
from glayout.components.stacked_current_mirror import stacked_nfet_current_mirror
from glayout.components.differential_to_single_ended_converter import differential_to_single_ended_converter
from glayout.components.row_csamplifier_diff_to_single_ended_converter import row_csamplifier_diff_to_single_ended_converter


@validate_arguments
def __add_diff_pair_and_bias(pdk: MappedPDK, toplevel_stacked: Component, half_diffpair_params: tuple[float, float, int], diffpair_bias: tuple[float, float, int], rmult: int, with_antenna_diode_on_diffinputs: int) -> Component:
    clear_cache()
    diffpair_i_ref = diff_pair_ibias(pdk, half_diffpair_params, diffpair_bias, rmult, with_antenna_diode_on_diffinputs)
    toplevel_stacked.add(diffpair_i_ref)
    toplevel_stacked.add_ports(diffpair_i_ref.get_ports_list(),prefix="diffpair_")

    toplevel_stacked.info['netlist'] = diffpair_i_ref.info['netlist']

    return toplevel_stacked

@validate_arguments
def __add_common_source_nbias_transistors(pdk: MappedPDK, toplevel_stacked: Component, half_common_source_nbias: tuple[float, float, int, int], rmult: int) -> Component:
    clear_cache()
    x_dim_center = toplevel_stacked.xmax
    for i in range(2):
        direction = (-1) ** i
        cmirrorref_ref, cmirrorout_ref = stacked_nfet_current_mirror(pdk, half_common_source_nbias, rmult, direction < 0)
        # xtranslation
        xtranslationO = direction * abs(x_dim_center + cmirrorout_ref.xmax + pdk.util_max_metal_seperation())
        xtranslationR = direction * abs(x_dim_center + cmirrorref_ref.xmax + pdk.util_max_metal_seperation())
        xtranslationO, xtranslationR = pdk.snap_to_2xgrid([xtranslationO, xtranslationR])
        cmirrorout_ref.movex(xtranslationO)
        cmirrorref_ref.movex(xtranslationR)
        # ytranslation
        cmirrorout_ref.movey(toplevel_stacked.ports["diffpair_bl_multiplier_0_gate_S"].center[1])
        cmirrorref_ref.movey(cmirrorout_ref.ymin - evaluate_bbox(cmirrorref_ref)[1]/2 - pdk.util_max_metal_seperation())
        # add ports
        toplevel_stacked.add(cmirrorref_ref)
        toplevel_stacked.add(cmirrorout_ref)
        side = "R" if i==0 else "L"
        toplevel_stacked.add_ports(cmirrorout_ref.get_ports_list(), prefix="commonsource_cmirror_output_"+side+"_")
        toplevel_stacked.add_ports(cmirrorref_ref.get_ports_list(), prefix="commonsource_cmirror_ref_"+side+"_")
        toplevel_stacked << straight_route(pdk, toplevel_stacked.ports["commonsource_cmirror_output_"+side+"_tie_S_top_met_S"], toplevel_stacked.ports["commonsource_cmirror_ref_"+side+"_tie_N_top_met_N"])
    return toplevel_stacked

@validate_arguments
def __route_bottom_ncomps_except_drain_nbias(pdk: MappedPDK, toplevel_stacked: Component, gndpin: Union[Component,ComponentReference], halfmultn_num_mults: int) -> tuple:
    clear_cache()
    # route diff pair cmirror
    toplevel_stacked << L_route(pdk, toplevel_stacked.ports["diffpair_ibias_purposegndport"],gndpin.ports["W"])
    # gnd diff pair substrate tap
    toplevel_stacked << straight_route(pdk, toplevel_stacked.ports["diffpair_tap_W_top_met_E"], toplevel_stacked.ports["commonsource_cmirror_output_L_tie_E_top_met_W"],width=1,glayer2="met1")
    toplevel_stacked << straight_route(pdk, toplevel_stacked.ports["diffpair_tap_E_top_met_W"], toplevel_stacked.ports["commonsource_cmirror_output_R_tie_W_top_met_E"],width=1,glayer2="met1")
    # common source
    # route to gnd the sources of cmirror
    _cref = toplevel_stacked << c_route(pdk, toplevel_stacked.ports["commonsource_cmirror_output_R_multiplier_0_source_con_S"], toplevel_stacked.ports["commonsource_cmirror_output_L_multiplier_0_source_con_S"], extension=abs(gndpin.ports["N"].center[1]-toplevel_stacked.ports["commonsource_cmirror_output_R_multiplier_0_source_con_S"].center[1]),fullbottom=True)
    toplevel_stacked << straight_route(pdk, toplevel_stacked.ports["commonsource_cmirror_ref_R_multiplier_0_source_E"],_cref.ports["con_E"],glayer2="met3",via2_alignment=('c','c'))
    toplevel_stacked << straight_route(pdk, toplevel_stacked.ports["commonsource_cmirror_ref_L_multiplier_0_source_W"],_cref.ports["con_W"],glayer2="met3",via2_alignment=('c','c'))
    # connect cmirror ref drain to cmirror output gate, then short cmirror ref drain and gate
    Ldrainport = toplevel_stacked.ports["commonsource_cmirror_ref_L_multiplier_0_drain_N"]
    Lgateport = toplevel_stacked.ports["commonsource_cmirror_output_L_multiplier_0_gate_S"]
    Rdrainport = toplevel_stacked.ports["commonsource_cmirror_ref_R_multiplier_0_drain_N"]
    Rgateport = toplevel_stacked.ports["commonsource_cmirror_output_R_multiplier_0_gate_S"]
    draintogate_L = toplevel_stacked << straight_route(pdk, Ldrainport, Lgateport, glayer1="met3",via1_alignment=('c','b'),via2_alignment=('c','t'),width=1)
    draintogate_R = toplevel_stacked << straight_route(pdk, Rdrainport, Rgateport, glayer1="met3",via1_alignment=('c','b'),via2_alignment=('c','t'),width=1)
    Lcmirrorrefgate = toplevel_stacked.ports["commonsource_cmirror_ref_L_multiplier_0_gate_E"]
    Rcmirrorrefgate = toplevel_stacked.ports["commonsource_cmirror_ref_R_multiplier_0_gate_W"]
    extension = pdk.util_max_metal_seperation()
    toplevel_stacked << c_route(pdk, toplevel_stacked.ports["commonsource_cmirror_ref_L_multiplier_0_drain_E"], Lcmirrorrefgate, extension=extension)
    toplevel_stacked << c_route(pdk, toplevel_stacked.ports["commonsource_cmirror_ref_R_multiplier_0_drain_W"], Rcmirrorrefgate, extension=extension)
    # connect gates and drains of cmirror output
    halfMultn_left_gate_port = toplevel_stacked.ports["commonsource_cmirror_output_R_multiplier_"+str(halfmultn_num_mults-2)+"_gate_con_N"]
    halfMultn_right_gate_port = toplevel_stacked.ports["commonsource_cmirror_output_L_multiplier_"+str(halfmultn_num_mults-2)+"_gate_con_N"]
    halfmultn_gate_routeref = toplevel_stacked << c_route(pdk, halfMultn_left_gate_port, halfMultn_right_gate_port, extension=abs(toplevel_stacked.ymax-halfMultn_left_gate_port.center[1])+1,fullbottom=True, viaoffset=(False,False))
    halfMultn_left_drain_port = toplevel_stacked.ports["commonsource_cmirror_output_R_multiplier_"+str(halfmultn_num_mults-2)+"_drain_con_N"]
    halfMultn_right_drain_port = toplevel_stacked.ports["commonsource_cmirror_output_L_multiplier_"+str(halfmultn_num_mults-2)+"_drain_con_N"]
    halfmultn_drain_routeref = toplevel_stacked << c_route(pdk, halfMultn_left_drain_port, halfMultn_right_drain_port, extension=abs(toplevel_stacked.ymax-halfMultn_left_drain_port.center[1])+1,fullbottom=True)
    # route to gnd the guardring of cmirror output and the diff pair cmirror ring
    toplevel_stacked << straight_route(pdk,toplevel_stacked.ports["commonsource_cmirror_ref_R_tie_S_top_met_S"],movey(gndpin.ports["W"],evaluate_bbox(gndpin)[1]/4),width=2,glayer1="met3",fullbottom=True)
    toplevel_stacked << straight_route(pdk,toplevel_stacked.ports["commonsource_cmirror_ref_L_tie_S_top_met_S"],movey(gndpin.ports["E"],evaluate_bbox(gndpin)[1]/4),width=2,glayer1="met3",fullbottom=True)
    toplevel_stacked << straight_route(pdk,toplevel_stacked.ports["commonsource_cmirror_ref_L_tie_E_top_met_E"],toplevel_stacked.ports["diffpair_ibias_welltie_W_top_met_W"])
    toplevel_stacked << straight_route(pdk,toplevel_stacked.ports["commonsource_cmirror_ref_R_tie_W_top_met_W"],toplevel_stacked.ports["diffpair_ibias_welltie_E_top_met_E"])
    # diffpair
    # route source of diffpair to drain of diffpair cmirror
    toplevel_stacked << L_route(pdk,toplevel_stacked.ports["diffpair_source_routeW_con_N"],toplevel_stacked.ports["diffpair_ibias_B_drain_W"])
    toplevel_stacked << L_route(pdk,toplevel_stacked.ports["diffpair_source_routeE_con_N"],toplevel_stacked.ports["diffpair_ibias_B_drain_E"])
    return toplevel_stacked, halfmultn_drain_routeref, halfmultn_gate_routeref, _cref


def diff_pair_stackedcmirror(
    pdk: MappedPDK,
    half_diffpair_params: tuple[float, float, int],
    diffpair_bias: tuple[float, float, int],
    half_common_source_nbias: tuple[float, float, int, int],
    rmult: int,
    with_antenna_diode_on_diffinputs: int
) -> Component:
    # create toplevel_stacked component
    toplevel_stacked = Component()
    # place nmos components
    diffpair_and_bias = __add_diff_pair_and_bias(pdk, toplevel_stacked, half_diffpair_params, diffpair_bias, rmult, with_antenna_diode_on_diffinputs)
    # create and position each half of the nmos bias transistor for the common source stage symetrically
    toplevel_stacked = __add_common_source_nbias_transistors(pdk, toplevel_stacked, half_common_source_nbias, rmult)
    toplevel_stacked.add_padding(layers=(pdk.get_glayer("pwell"),),default=0)
    # add ground pin
    gndpin = toplevel_stacked << rename_ports_by_orientation(rectangle(size=(5,3),layer=pdk.get_glayer("met4"),centered=True))
    gndpin.movey(pdk.snap_to_2xgrid(toplevel_stacked.ymin-pdk.util_max_metal_seperation()-gndpin.ymax))
    # route bottom ncomps except drain of nbias (still need to place common source pmos amp)
    toplevel_stacked, halfmultn_drain_routeref, halfmultn_gate_routeref, _cref = __route_bottom_ncomps_except_drain_nbias(pdk, toplevel_stacked, gndpin, half_common_source_nbias[3])
    toplevel_stacked.add_ports(gndpin.get_ports_list(), prefix="pin_gnd_")

    return toplevel_stacked, halfmultn_drain_routeref, halfmultn_gate_routeref, _cref



from ..mappedpdk import MappedPDK

grulesobj = dict()
for glayer in MappedPDK.valid_glayers:
    grulesobj[glayer] = dict((x, None) for x in MappedPDK.valid_glayers)

grulesobj["dnwell"]["dnwell"] = {'min_width': 1.7, 'min_separation': 5.42}
grulesobj["dnwell"]["pwell"] = {'min_enclosure': 2.5}
grulesobj["dnwell"]["nwell"] = {'min_separation': 3.1, 'min_enclosure': 0.5}
grulesobj["dnwell"]["p+s/d"] = {}
grulesobj["dnwell"]["n+s/d"] = {}
grulesobj["dnwell"]["active_diff"] = {'min_enclosure': 0.93}
grulesobj["dnwell"]["active_tap"] = {'min_enclosure': 0.62, 'min_separation': 2.5}
grulesobj["dnwell"]["poly"] = {}
grulesobj["dnwell"]["mcon"] = {}
grulesobj["dnwell"]["met1"] = {}
grulesobj["dnwell"]["via1"] = {}
grulesobj["dnwell"]["met2"] = {}
grulesobj["dnwell"]["via2"] = {}
grulesobj["dnwell"]["met3"] = {}
grulesobj["dnwell"]["via3"] = {}
grulesobj["dnwell"]["met4"] = {}
grulesobj["dnwell"]["via4"] = {}
grulesobj["dnwell"]["met5"] = {}
grulesobj["dnwell"]["capmet"] = {}
grulesobj["pwell"]["dnwell"] = {}
grulesobj["pwell"]["pwell"] = {'min_width': 0.6, 'min_separation': 1.4}
grulesobj["pwell"]["nwell"] = {'min_separation': 0.0}
grulesobj["pwell"]["p+s/d"] = {}
grulesobj["pwell"]["n+s/d"] = {}
grulesobj["pwell"]["active_diff"] = {'min_enclosure': 0.43}
grulesobj["pwell"]["active_tap"] = {'min_enclosure': 0.12}
grulesobj["pwell"]["poly"] = {}
grulesobj["pwell"]["mcon"] = {}
grulesobj["pwell"]["met1"] = {}
grulesobj["pwell"]["via1"] = {}
grulesobj["pwell"]["met2"] = {}
grulesobj["pwell"]["via2"] = {}
grulesobj["pwell"]["met3"] = {}
grulesobj["pwell"]["via3"] = {}
grulesobj["pwell"]["met4"] = {}
grulesobj["pwell"]["via4"] = {}
grulesobj["pwell"]["met5"] = {}
grulesobj["pwell"]["capmet"] = {}
grulesobj["nwell"]["dnwell"] = {}
grulesobj["nwell"]["pwell"] = {}
grulesobj["nwell"]["nwell"] = {'min_width': 0.86, 'min_separation': 1.4}
grulesobj["nwell"]["p+s/d"] = {}
grulesobj["nwell"]["n+s/d"] = {}
grulesobj["nwell"]["active_diff"] = {'min_enclosure': 0.43}
grulesobj["nwell"]["active_tap"] = {'min_enclosure': 0.12}
grulesobj["nwell"]["poly"] = {}
grulesobj["nwell"]["mcon"] = {}
grulesobj["nwell"]["met1"] = {}
grulesobj["nwell"]["via1"] = {}
grulesobj["nwell"]["met2"] = {}
grulesobj["nwell"]["via2"] = {}
grulesobj["nwell"]["met3"] = {}
grulesobj["nwell"]["via3"] = {}
grulesobj["nwell"]["met4"] = {}
grulesobj["nwell"]["via4"] = {}
grulesobj["nwell"]["met5"] = {}
grulesobj["nwell"]["capmet"] = {}
grulesobj["p+s/d"]["dnwell"] = {}
grulesobj["p+s/d"]["pwell"] = {}
grulesobj["p+s/d"]["nwell"] = {}
grulesobj["p+s/d"]["p+s/d"] = {'min_width': 0.4, 'min_separation': 0.4}
grulesobj["p+s/d"]["n+s/d"] = {}
grulesobj["p+s/d"]["active_diff"] = {'min_enclosure': 0.23}
grulesobj["p+s/d"]["active_tap"] = {'min_enclosure': 0.16}
grulesobj["p+s/d"]["poly"] = {}
grulesobj["p+s/d"]["mcon"] = {}
grulesobj["p+s/d"]["met1"] = {}
grulesobj["p+s/d"]["via1"] = {}
grulesobj["p+s/d"]["met2"] = {}
grulesobj["p+s/d"]["via2"] = {}
grulesobj["p+s/d"]["met3"] = {}
grulesobj["p+s/d"]["via3"] = {}
grulesobj["p+s/d"]["met4"] = {}
grulesobj["p+s/d"]["via4"] = {}
grulesobj["p+s/d"]["met5"] = {}
grulesobj["p+s/d"]["capmet"] = {}
grulesobj["n+s/d"]["dnwell"] = {}
grulesobj["n+s/d"]["pwell"] = {}
grulesobj["n+s/d"]["nwell"] = {}
grulesobj["n+s/d"]["p+s/d"] = {}
grulesobj["n+s/d"]["n+s/d"] = {'min_width': 0.4, 'min_separation': 0.4}
grulesobj["n+s/d"]["active_diff"] = {'min_enclosure': 0.23}
grulesobj["n+s/d"]["active_tap"] = {'min_enclosure': 0.16}
grulesobj["n+s/d"]["poly"] = {}
grulesobj["n+s/d"]["mcon"] = {}
grulesobj["n+s/d"]["met1"] = {}
grulesobj["n+s/d"]["via1"] = {}
grulesobj["n+s/d"]["met2"] = {}
grulesobj["n+s/d"]["via2"] = {}
grulesobj["n+s/d"]["met3"] = {}
grulesobj["n+s/d"]["via3"] = {}
grulesobj["n+s/d"]["met4"] = {}
grulesobj["n+s/d"]["via4"] = {}
grulesobj["n+s/d"]["met5"] = {}
grulesobj["n+s/d"]["capmet"] = {}
grulesobj["active_diff"]["dnwell"] = {}
grulesobj["active_diff"]["pwell"] = {}
grulesobj["active_diff"]["nwell"] = {}
grulesobj["active_diff"]["p+s/d"] = {}
grulesobj["active_diff"]["n+s/d"] = {}
grulesobj["active_diff"]["active_diff"] = {'min_width': 0.22, 'min_separation': 0.28}
grulesobj["active_diff"]["active_tap"] = {'min_separation': 0.28, 'max_separation': 20.0}
grulesobj["active_diff"]["poly"] = {'overhang': 0.24, 'min_separation': 0.1}
grulesobj["active_diff"]["mcon"] = {'min_enclosure': 0.07}
grulesobj["active_diff"]["met1"] = {}
grulesobj["active_diff"]["via1"] = {}
grulesobj["active_diff"]["met2"] = {}
grulesobj["active_diff"]["via2"] = {}
grulesobj["active_diff"]["met3"] = {}
grulesobj["active_diff"]["via3"] = {}
grulesobj["active_diff"]["met4"] = {}
grulesobj["active_diff"]["via4"] = {}
grulesobj["active_diff"]["met5"] = {}
grulesobj["active_diff"]["capmet"] = {}
grulesobj["active_tap"]["dnwell"] = {}
grulesobj["active_tap"]["pwell"] = {}
grulesobj["active_tap"]["nwell"] = {}
grulesobj["active_tap"]["p+s/d"] = {}
grulesobj["active_tap"]["n+s/d"] = {}
grulesobj["active_tap"]["active_diff"] = {}
grulesobj["active_tap"]["active_tap"] = {'min_width': 0.22, 'min_separation': 0.28}
grulesobj["active_tap"]["poly"] = {'min_separation': 0.1}
grulesobj["active_tap"]["mcon"] = {'min_enclosure': 0.07}
grulesobj["active_tap"]["met1"] = {}
grulesobj["active_tap"]["via1"] = {}
grulesobj["active_tap"]["met2"] = {}
grulesobj["active_tap"]["via2"] = {}
grulesobj["active_tap"]["met3"] = {}
grulesobj["active_tap"]["via3"] = {}
grulesobj["active_tap"]["met4"] = {}
grulesobj["active_tap"]["via4"] = {}
grulesobj["active_tap"]["met5"] = {}
grulesobj["active_tap"]["capmet"] = {}
grulesobj["poly"]["dnwell"] = {}
grulesobj["poly"]["pwell"] = {}
grulesobj["poly"]["nwell"] = {}
grulesobj["poly"]["p+s/d"] = {}
grulesobj["poly"]["n+s/d"] = {}
grulesobj["poly"]["active_diff"] = {}
grulesobj["poly"]["active_tap"] = {}
grulesobj["poly"]["poly"] = {'min_width': 0.28}
grulesobj["poly"]["mcon"] = {'min_enclosure': 0.07, 'min_separation': 0.17}
grulesobj["poly"]["met1"] = {}
grulesobj["poly"]["via1"] = {}
grulesobj["poly"]["met2"] = {}
grulesobj["poly"]["via2"] = {}
grulesobj["poly"]["met3"] = {}
grulesobj["poly"]["via3"] = {}
grulesobj["poly"]["met4"] = {}
grulesobj["poly"]["via4"] = {}
grulesobj["poly"]["met5"] = {}
grulesobj["poly"]["capmet"] = {}
grulesobj["mcon"]["dnwell"] = {}
grulesobj["mcon"]["pwell"] = {}
grulesobj["mcon"]["nwell"] = {}
grulesobj["mcon"]["p+s/d"] = {}
grulesobj["mcon"]["n+s/d"] = {}
grulesobj["mcon"]["active_diff"] = {}
grulesobj["mcon"]["active_tap"] = {}
grulesobj["mcon"]["poly"] = {}
grulesobj["mcon"]["mcon"] = {'min_separation': 0.28, 'width': 0.22}
grulesobj["mcon"]["met1"] = {'min_enclosure': 0.12}
grulesobj["mcon"]["via1"] = {}
grulesobj["mcon"]["met2"] = {}
grulesobj["mcon"]["via2"] = {}
grulesobj["mcon"]["met3"] = {}
grulesobj["mcon"]["via3"] = {}
grulesobj["mcon"]["met4"] = {}
grulesobj["mcon"]["via4"] = {}
grulesobj["mcon"]["met5"] = {}
grulesobj["mcon"]["capmet"] = {}
grulesobj["met1"]["dnwell"] = {}
grulesobj["met1"]["pwell"] = {}
grulesobj["met1"]["nwell"] = {}
grulesobj["met1"]["p+s/d"] = {}
grulesobj["met1"]["n+s/d"] = {}
grulesobj["met1"]["active_diff"] = {}
grulesobj["met1"]["active_tap"] = {}
grulesobj["met1"]["poly"] = {}
grulesobj["met1"]["mcon"] = {}
grulesobj["met1"]["met1"] = {'min_width': 0.23, 'min_separation': 0.3}
grulesobj["met1"]["via1"] = {'min_enclosure': 0.12}
grulesobj["met1"]["met2"] = {}
grulesobj["met1"]["via2"] = {}
grulesobj["met1"]["met3"] = {}
grulesobj["met1"]["via3"] = {}
grulesobj["met1"]["met4"] = {}
grulesobj["met1"]["via4"] = {}
grulesobj["met1"]["met5"] = {}
grulesobj["met1"]["capmet"] = {}
grulesobj["via1"]["dnwell"] = {}
grulesobj["via1"]["pwell"] = {}
grulesobj["via1"]["nwell"] = {}
grulesobj["via1"]["p+s/d"] = {}
grulesobj["via1"]["n+s/d"] = {}
grulesobj["via1"]["active_diff"] = {}
grulesobj["via1"]["active_tap"] = {}
grulesobj["via1"]["poly"] = {}
grulesobj["via1"]["mcon"] = {}
grulesobj["via1"]["met1"] = {}
grulesobj["via1"]["via1"] = {'width': 0.26, 'min_separation': 0.36}
grulesobj["via1"]["met2"] = {'min_enclosure': 0.12}
grulesobj["via1"]["via2"] = {}
grulesobj["via1"]["met3"] = {}
grulesobj["via1"]["via3"] = {}
grulesobj["via1"]["met4"] = {}
grulesobj["via1"]["via4"] = {}
grulesobj["via1"]["met5"] = {}
grulesobj["via1"]["capmet"] = {}
grulesobj["met2"]["dnwell"] = {}
grulesobj["met2"]["pwell"] = {}
grulesobj["met2"]["nwell"] = {}
grulesobj["met2"]["p+s/d"] = {}
grulesobj["met2"]["n+s/d"] = {}
grulesobj["met2"]["active_diff"] = {}
grulesobj["met2"]["active_tap"] = {}
grulesobj["met2"]["poly"] = {}
grulesobj["met2"]["mcon"] = {}
grulesobj["met2"]["met1"] = {}
grulesobj["met2"]["via1"] = {}
grulesobj["met2"]["met2"] = {'min_width': 0.28, 'min_separation': 0.3}
grulesobj["met2"]["via2"] = {'min_enclosure': 0.12}
grulesobj["met2"]["met3"] = {}
grulesobj["met2"]["via3"] = {}
grulesobj["met2"]["met4"] = {}
grulesobj["met2"]["via4"] = {}
grulesobj["met2"]["met5"] = {}
grulesobj["met2"]["capmet"] = {'min_enclosure': 0.6}
grulesobj["via2"]["dnwell"] = {}
grulesobj["via2"]["pwell"] = {}
grulesobj["via2"]["nwell"] = {}
grulesobj["via2"]["p+s/d"] = {}
grulesobj["via2"]["n+s/d"] = {}
grulesobj["via2"]["active_diff"] = {}
grulesobj["via2"]["active_tap"] = {}
grulesobj["via2"]["poly"] = {}
grulesobj["via2"]["mcon"] = {}
grulesobj["via2"]["met1"] = {}
grulesobj["via2"]["via1"] = {}
grulesobj["via2"]["met2"] = {}
grulesobj["via2"]["via2"] = {'width': 0.26, 'min_separation': 0.36}
grulesobj["via2"]["met3"] = {'min_enclosure': 0.12}
grulesobj["via2"]["via3"] = {}
grulesobj["via2"]["met4"] = {}
grulesobj["via2"]["via4"] = {}
grulesobj["via2"]["met5"] = {}
grulesobj["via2"]["capmet"] = {}
grulesobj["met3"]["dnwell"] = {}
grulesobj["met3"]["pwell"] = {}
grulesobj["met3"]["nwell"] = {}
grulesobj["met3"]["p+s/d"] = {}
grulesobj["met3"]["n+s/d"] = {}
grulesobj["met3"]["active_diff"] = {}
grulesobj["met3"]["active_tap"] = {}
grulesobj["met3"]["poly"] = {}
grulesobj["met3"]["mcon"] = {}
grulesobj["met3"]["met1"] = {}
grulesobj["met3"]["via1"] = {}
grulesobj["met3"]["met2"] = {}
grulesobj["met3"]["via2"] = {}
grulesobj["met3"]["met3"] = {'min_width': 0.28, 'min_separation': 0.3}
grulesobj["met3"]["via3"] = {'min_enclosure': 0.12}
grulesobj["met3"]["met4"] = {}
grulesobj["met3"]["via4"] = {}
grulesobj["met3"]["met5"] = {}
grulesobj["met3"]["capmet"] = {}
grulesobj["via3"]["dnwell"] = {}
grulesobj["via3"]["pwell"] = {}
grulesobj["via3"]["nwell"] = {}
grulesobj["via3"]["p+s/d"] = {}
grulesobj["via3"]["n+s/d"] = {}
grulesobj["via3"]["active_diff"] = {}
grulesobj["via3"]["active_tap"] = {}
grulesobj["via3"]["poly"] = {}
grulesobj["via3"]["mcon"] = {}
grulesobj["via3"]["met1"] = {}
grulesobj["via3"]["via1"] = {}
grulesobj["via3"]["met2"] = {}
grulesobj["via3"]["via2"] = {}
grulesobj["via3"]["met3"] = {}
grulesobj["via3"]["via3"] = {'width': 0.26, 'min_separation': 0.36}
grulesobj["via3"]["met4"] = {'min_enclosure': 0.12}
grulesobj["via3"]["via4"] = {}
grulesobj["via3"]["met5"] = {}
grulesobj["via3"]["capmet"] = {}
grulesobj["met4"]["dnwell"] = {}
grulesobj["met4"]["pwell"] = {}
grulesobj["met4"]["nwell"] = {}
grulesobj["met4"]["p+s/d"] = {}
grulesobj["met4"]["n+s/d"] = {}
grulesobj["met4"]["active_diff"] = {}
grulesobj["met4"]["active_tap"] = {}
grulesobj["met4"]["poly"] = {}
grulesobj["met4"]["mcon"] = {}
grulesobj["met4"]["met1"] = {}
grulesobj["met4"]["via1"] = {}
grulesobj["met4"]["met2"] = {}
grulesobj["met4"]["via2"] = {}
grulesobj["met4"]["met3"] = {}
grulesobj["met4"]["via3"] = {}
grulesobj["met4"]["met4"] = {'min_width': 0.28, 'min_separation': 0.3}
grulesobj["met4"]["via4"] = {'min_enclosure': 0.12}
grulesobj["met4"]["met5"] = {}
grulesobj["met4"]["capmet"] = {}
grulesobj["via4"]["dnwell"] = {}
grulesobj["via4"]["pwell"] = {}
grulesobj["via4"]["nwell"] = {}
grulesobj["via4"]["p+s/d"] = {}
grulesobj["via4"]["n+s/d"] = {}
grulesobj["via4"]["active_diff"] = {}
grulesobj["via4"]["active_tap"] = {}
grulesobj["via4"]["poly"] = {}
grulesobj["via4"]["mcon"] = {}
grulesobj["via4"]["met1"] = {}
grulesobj["via4"]["via1"] = {}
grulesobj["via4"]["met2"] = {}
grulesobj["via4"]["via2"] = {}
grulesobj["via4"]["met3"] = {}
grulesobj["via4"]["via3"] = {}
grulesobj["via4"]["met4"] = {}
grulesobj["via4"]["via4"] = {'width': 0.26, 'min_separation': 0.36}
grulesobj["via4"]["met5"] = {'min_enclosure': 0.12}
grulesobj["via4"]["capmet"] = {}
grulesobj["met5"]["dnwell"] = {}
grulesobj["met5"]["pwell"] = {}
grulesobj["met5"]["nwell"] = {}
grulesobj["met5"]["p+s/d"] = {}
grulesobj["met5"]["n+s/d"] = {}
grulesobj["met5"]["active_diff"] = {}
grulesobj["met5"]["active_tap"] = {}
grulesobj["met5"]["poly"] = {}
grulesobj["met5"]["mcon"] = {}
grulesobj["met5"]["met1"] = {}
grulesobj["met5"]["via1"] = {}
grulesobj["met5"]["met2"] = {}
grulesobj["met5"]["via2"] = {}
grulesobj["met5"]["met3"] = {}
grulesobj["met5"]["via3"] = {}
grulesobj["met5"]["met4"] = {}
grulesobj["met5"]["via4"] = {}
grulesobj["met5"]["met5"] = {'min_width': 0.28, 'min_separation': 0.3}
grulesobj["met5"]["capmet"] = {}
grulesobj["capmet"]["dnwell"] = {}
grulesobj["capmet"]["pwell"] = {}
grulesobj["capmet"]["nwell"] = {}
grulesobj["capmet"]["p+s/d"] = {}
grulesobj["capmet"]["n+s/d"] = {}
grulesobj["capmet"]["active_diff"] = {}
grulesobj["capmet"]["active_tap"] = {}
grulesobj["capmet"]["poly"] = {}
grulesobj["capmet"]["mcon"] = {}
grulesobj["capmet"]["met1"] = {}
grulesobj["capmet"]["via1"] = {}
grulesobj["capmet"]["met2"] = {}
grulesobj["capmet"]["via2"] = {}
grulesobj["capmet"]["met3"] = {}
grulesobj["capmet"]["via3"] = {}
grulesobj["capmet"]["met4"] = {}
grulesobj["capmet"]["via4"] = {}
grulesobj["capmet"]["met5"] = {}
grulesobj["capmet"]["capmet"] = {'capmettop': (42, 0), 'capmetbottom': (36, 0), 'min_separation': 1.2}




"""
usage: from sky130_mapped import sky130_mapped_pdk
"""

from ..mappedpdk import MappedPDK
from ..sky130_mapped.grules import grulesobj
from pathlib import Path
from ..sky130_mapped.sky130_add_npc import sky130_add_npc


# use mimcap over metal 3
sky130_glayer_mapping = {
    "capmet": (89, 44),
    "met5": (71,20),
    "via4": (70,44),
    "met4": (70,20),
    "via3": (69,44),
    "met3": (69,20),
    "via2": (68,44),
    "met2": (68,20),
    "via1": (67,44),
    "met1": (67,20),
    "mcon": (66,44),
    "poly": (66,20),
    "active_diff": (65,20),
    "active_tap": (65,20),
    "n+s/d": (93,44),
    "p+s/d": (94,20),
    "nwell": (64,20),
    "pwell": (64,44),
    "dnwell": (64,18),
}


sky130_lydrc_file_path = Path(__file__).resolve().parent / "sky130.lydrc"


sky130_mapped_pdk = MappedPDK(
    name="sky130",
    glayers=sky130_glayer_mapping,
	models={
        'nfet': 'sky130_fd_pr__nfet_01v8',
		'pfet': 'sky130_fd_pr__pfet_01v8',
		'mimcap': 'sky130_fd_pr__cap_mim_m3_1'
    },
    grules=grulesobj,
    #klayout_lydrc_file=sky130_lydrc_file_path,
    default_decorator=sky130_add_npc
)
# set the grid size
sky130_mapped_pdk.gds_write_settings.precision = 5*10**-9
sky130_mapped_pdk.cell_decorator_settings.cache=False
sky130_mapped_pdk.gds_write_settings.flatten_invalid_refs=False



from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.rectangle import rectangle
from pydantic import validate_arguments
from glayout.pdk.mappedpdk import MappedPDK
from math import floor
from typing import Optional, Union
from glayout.pdk.util.comp_utils import evaluate_bbox, prec_array, to_float, move, prec_ref_center, to_decimal
from glayout.pdk.util.port_utils import rename_ports_by_orientation, print_ports
from glayout.pdk.util.snap_to_grid import component_snap_to_grid
from decimal import Decimal
from typing import Literal


@validate_arguments
def __error_check_order_layers(
    pdk: MappedPDK, glayer1: str, glayer2: str
) -> tuple[tuple[int, int], tuple[str, str]]:
    """correctly order layers (level1 should be lower than level2)"""
    pdk.activate()
    # check that the generic layers specfied can be routed between
    if not all([pdk.is_routable_glayer(met) for met in [glayer1, glayer2]]):
        raise ValueError("via_stack: specify between two routable layers")
    level1 = int(glayer1[-1]) if "met" in glayer1 else 0
    level2 = int(glayer2[-1]) if "met" in glayer2 else 0
    if level1 > level2:
        level1, level2 = level2, level1
        glayer1, glayer2 = glayer2, glayer1
    # check that all layers needed between glayer1-glayer2 are present
    required_glayers = [glayer2]
    for level in range(level1,level2):
        via_name = "mcon" if level==0 else "via"+str(level)
        layer_name = glayer1 if level==0 else "met"+str(level)
        required_glayers += [via_name,layer_name]
    pdk.has_required_glayers(required_glayers)
    return ((level1,level2),(glayer1,glayer2))


@validate_arguments
def __get_layer_dim(pdk: MappedPDK, glayer: str, mode: Literal["both","above","below"]="both") -> float:
	"""Returns the required dimension of a routable layer in a via stack
	glayer is the routable glayer
	mode is one of [both,below,above]
	This specfies the vias to consider. (layer dims may be made smaller if its possible to ignore top/bottom vias)
	****enclosure rules of the via above and below are considered by default, via1<->met2<->via2
	****using below specfier only considers the enclosure rules for the via below, via1<->met2
	****using above specfier only considers the enclosure rules for the via above, met2<->via2
	****specfying both or below for active/poly layer is valid, function knows to ignore below
	"""
	# error checking
	if not pdk.is_routable_glayer(glayer):
		raise ValueError("__get_layer_dim: glayer must be a routable layer")
	# split into above rules and below rules
	consider_above = (mode=="both" or mode=="above")
	consider_below = (mode=="both" or mode=="below")
	is_lvl0 = any([hint in glayer for hint in ["poly","active"]])
	layer_dim=0
	if consider_below and not is_lvl0:
		via_below = "mcon" if glayer=="met1" else "via"+str(int(glayer[-1])-1)
		layer_dim = pdk.get_grule(via_below)["width"] + 2*pdk.get_grule(via_below,glayer)["min_enclosure"]
	if consider_above:
		via_above = "mcon" if is_lvl0 else "via"+str(glayer[-1])
		layer_dim = max(layer_dim, pdk.get_grule(via_above)["width"] + 2*pdk.get_grule(via_above,glayer)["min_enclosure"])
	layer_dim = max(layer_dim, pdk.get_grule(glayer)["min_width"])
	return layer_dim


@validate_arguments
def __get_viastack_minseperation(pdk: MappedPDK, viastack: Component, ordered_layer_info) -> tuple[float,float]:
    """internal use: return absolute via separation and top_enclosure (top via to top met enclosure)"""
    get_sep = lambda _pdk, rule, _lay_, comp : (rule+2*comp.extract(layers=[_pdk.get_glayer(_lay_)]).xmax)
    level1, level2 = ordered_layer_info[0]
    glayer1, glayer2 = ordered_layer_info[1]
    mcon_rule = pdk.get_grule("mcon")["min_separation"]
    via_spacing = [] if level1 else [get_sep(pdk,mcon_rule,"mcon",viastack)]
    level1_met = level1 if level1 else level1 + 1
    top_enclosure = 0
    for level in range(level1_met, level2):
        met_glayer = "met" + str(level)
        via_glayer = "via" + str(level)
        mrule = pdk.get_grule(met_glayer)["min_separation"]
        vrule = pdk.get_grule(via_glayer)["min_separation"]
        via_spacing.append(get_sep(pdk, mrule,met_glayer,viastack))
        via_spacing.append(get_sep(pdk, vrule,via_glayer,viastack))
        if level == (level2-1):
            top_enclosure = pdk.get_grule(glayer2,via_glayer)["min_enclosure"]
    via_spacing = pdk.snap_to_2xgrid(max(via_spacing),return_type="float")
    top_enclosure = pdk.snap_to_2xgrid(top_enclosure,return_type="float")
    return pdk.snap_to_2xgrid([via_spacing, 2*top_enclosure], return_type="float")


@cell
def via_stack(
    pdk: MappedPDK,
    glayer1: str,
    glayer2: str,
    centered: bool = True,
    fullbottom: bool = False,
    fulltop: bool = False,
    assume_bottom_via: bool = False,
    same_layer_behavior: Literal["lay_nothing","min_square"] = "lay_nothing"
) -> Component:
    """produces a single via stack between two layers that are routable (metal, poly, or active)
    The via_stack produced is always a square (hieght=width)
    
    args:
    pdk: MappedPDK is the pdk to use
    glayer1: str is the glayer to start on
    glayer2: str is the glayer to end on
    ****NOTE it does not matter what order you pass layers
    fullbottom: if True will lay the bottom layer all over the area of the viastack else makes minimum legal size (ignores min area)
    assume_bottom_via: legalize viastack assuming the via underneath bottom met is present, e.g. if bottom met is met3, assume via2 is present
    fulltop: if True will lay the top layer all over the area of the viastack else makes minimum legal size (ignores min area)
    ****NOTE: generator can figure out which layer is top and which is bottom (i.e. met5 is higher than met1)
    same_layer_behavior: sometimes (especially when used in other generators) it is unknown what two layers are specfied
    this option provides the generator with guidance on how to handle a case where same layer is given
    by default, (lay_nothing option) nothing is laid and an empty component is returned
    if min_square is specfied, a square of min_width * min_width is laid
    
    ports, some ports are not layed when it does not make sense (e.g. empty component):
    top_met_...all edges
    bottom_via_...all edges
    bottom_met_...all edges
    bottom_layer_...all edges (may be different than bottom met if on diff/poly)
    """
    ordered_layer_info = __error_check_order_layers(pdk, glayer1, glayer2)
    level1, level2 = ordered_layer_info[0]
    glayer1, glayer2 = ordered_layer_info[1]
    viastack = Component()
    # if same level return component with min_width rectangle on that layer
    if level1 == level2:
        if same_layer_behavior=="lay_nothing":
            return viastack
        min_square = viastack << rectangle(size=2*[pdk.get_grule(glayer1)["min_width"]],layer=pdk.get_glayer(glayer1), centered=centered)
        # update ports
        if level1==0:# both poly or active
            viastack.add_ports(min_square.get_ports_list(),prefix="bottom_layer_")
        else:# both mets
            viastack.add_ports(min_square.get_ports_list(),prefix="top_met_")
            viastack.add_ports(min_square.get_ports_list(),prefix="bottom_met_")
    else:
        ports_to_add = dict()
        for level in range(level1,level2+1):
            via_name = "mcon" if level==0 else "via"+str(level)
            layer_name = glayer1 if level==0 else "met"+str(level)
            # get layer sizing
            mode = "below" if level==level2 else ("above" if level==level1 else "both")
            mode = "both" if assume_bottom_via and level==level1 else mode
            layer_dim = __get_layer_dim(pdk, layer_name, mode=mode)
            # place met/via, do not place via if on top layer
            if level != level2:
                via_dim = pdk.get_grule(via_name)["width"]
                via_ref = viastack << rectangle(size=[via_dim,via_dim],layer=pdk.get_glayer(via_name), centered=True)
            lay_ref = viastack << rectangle(size=[layer_dim,layer_dim],layer=pdk.get_glayer(layer_name), centered=True)
            # update ports
            if layer_name == glayer1:
                ports_to_add["bottom_layer_"] = lay_ref.get_ports_list()
                ports_to_add["bottom_via_"] = via_ref.get_ports_list()
            if (level1==0 and level==1) or (level1>0 and layer_name==glayer1):
                ports_to_add["bottom_met_"] = lay_ref.get_ports_list()
            if layer_name == glayer2:
                ports_to_add["top_met_"] = lay_ref.get_ports_list()
        # implement fulltop and fullbottom options. update ports_to_add accordingly 
        if fullbottom:
            bot_ref = viastack << rectangle(size=evaluate_bbox(viastack),layer=pdk.get_glayer(glayer1), centered=True)
            if level1!=0:
                ports_to_add["bottom_met_"] = bot_ref.get_ports_list()
            ports_to_add["bottom_layer_"] = bot_ref.get_ports_list()
        if fulltop:
            ports_to_add["top_met_"] = (viastack << rectangle(size=evaluate_bbox(viastack),layer=pdk.get_glayer(glayer2), centered=True)).get_ports_list()
        # add all ports in ports_to_add
        for prefix, ports_list in ports_to_add.items():
            viastack.add_ports(ports_list,prefix=prefix)
        # move SW corner to 0,0 if centered=False
        if not centered:
            viastack = move(viastack,(viastack.xmax,viastack.ymax))
    return rename_ports_by_orientation(viastack.flatten())


@cell
def via_array(
    pdk: MappedPDK,
    glayer1: str,
    glayer2: str,
    size: Optional[tuple[Optional[float],Optional[float]]] = None,
    minus1: bool = False,
    num_vias: Optional[tuple[Optional[int],Optional[int]]] = None,
    lay_bottom: bool = True,
    fullbottom: bool = False,
    no_exception: bool = False,
    lay_every_layer: bool = False
) -> Component:
    """Fill a region with vias. Will automatically decide num rows and columns
    args:
    pdk: MappedPDK is the pdk to use
    glayer1: str is the glayer to start on
    glayer2: str is the glayer to end on
    lay_bottom: bool if true will lay bottom layer (by default only lays top layer)
    ****NOTE it does not matter what order you pass layers
    ****NOTE will lay bottom only over the minimum area required to make legal
    size: tuple is the (width, hieght) of the area to enclose
    ****NOTE: the size will be the dimensions of the top metal
    minus1: if true removes 1 via from rows/cols num vias 
    ****use if you want extra space at the edges of the array, does not apply to num_vias
    num_vias: number of rows/cols in the via array. Overrides size option
    ****NOTE: you can specify size for one dim and num_vias for another by setting one element to None 
    ****NOTE: num_vias overides size option
    fullbottom: True specifies that the bottom layer should extend over the entire via_array region
    ****NOTE: fullbottom=True implies lay_bottom and overrides if False
    no_exception: True specfies that the function should change size such that min size is met
    lay_every_layer: True specifies that every layer between glayer1 and glayer2 should be layed in full (not just the vias).
    ****NOTE: this implies lay_bottom
    
    ports, some ports are not layed when it does not make sense (e.g. empty component):
    top_met_...all edges
    bottom_lay_...all edges (only if lay_bottom is specified)
    array_...all ports associated with via array
    """
    # setup
    ordered_layer_info = __error_check_order_layers(pdk, glayer1, glayer2)
    level1, level2 = ordered_layer_info[0]
    glayer1, glayer2 = ordered_layer_info[1]
    viaarray = Component()
    # if same level return empty component
    if level1 == level2:
        return viaarray
    # figure out min space between via stacks
    viastack = via_stack(pdk, glayer1, glayer2)
    viadim = evaluate_bbox(viastack)[0]
    via_abs_spacing, top_enclosure = __get_viastack_minseperation(pdk, viastack, ordered_layer_info)
    # error check size and determine num_vias, cnum_vias[0]=x, cnum_vias[1]=y
    cnum_vias = 2*[None]
    for i in range(2):
        if (num_vias[i] if num_vias else False):
            cnum_vias[i] = num_vias[i]
        elif (size[i] if size else False):
            dim = pdk.snap_to_2xgrid(size[i],return_type="float")
            fltnum = floor((dim - top_enclosure) / (via_abs_spacing)) or 1
            fltnum = 1 if fltnum < 1 else fltnum
            cnum_vias[i] = ((fltnum - 1) or 1) if minus1 else fltnum
            if to_decimal(viadim) > to_decimal(dim) and not no_exception:
                raise ValueError(f"via_array,size:dim#{i}={dim} < {viadim}")
        else:
            raise ValueError("give at least 1: num_vias or size for each dim")
    # create array
    viaarray_ref = prec_ref_center(prec_array(viastack, columns=cnum_vias[0], rows=cnum_vias[1], spacing=2*[via_abs_spacing],absolute_spacing=True))
    viaarray.add(viaarray_ref)
    viaarray.add_ports(viaarray_ref.get_ports_list(),prefix="array_")
    # find the what should be used as full dims
    viadims = evaluate_bbox(viaarray)
    if not size:
        size = 2*[None]
    size = [size[i] if size[i] else viadims[i] for i in range(2)]
    size = [viadims[i] if viadims[i]>size[i] else size[i] for i in range(2)]
    # place bottom layer and add bot_lay_ ports
    if lay_bottom or fullbottom or lay_every_layer:
        bdims = evaluate_bbox(viaarray.extract(layers=[pdk.get_glayer(glayer1)]))
        bref = viaarray << rectangle(size=(size if fullbottom else bdims), layer=pdk.get_glayer(glayer1), centered=True)
        viaarray.add_ports(bref.get_ports_list(), prefix="bottom_lay_")
    else:
        viaarray = viaarray.remove_layers(layers=[pdk.get_glayer(glayer1)])
    # place top met
    tref = viaarray << rectangle(size=size, layer=pdk.get_glayer(glayer2), centered=True)
    viaarray.add_ports(tref.get_ports_list(), prefix="top_met_")
    # place every layer in between if lay_every_layer
    if lay_every_layer:
        for i in range(level1+1,level2):
            bdims = evaluate_bbox(viaarray.extract(layers=[pdk.get_glayer(f"met{i}")]))
            viaarray << rectangle(size=bdims, layer=pdk.get_glayer(f"met{i}"), centered=True)
    return component_snap_to_grid(rename_ports_by_orientation(viaarray))





from gdsfactory.component import Component
from gdsfactory.polygon import Polygon
from gdsfactory.geometry.boolean import boolean


def sky130_add_npc(comp: Component) -> Component:
	"""To keep with the generic generator structure,
	we do NOT add nitride poly cut layer in the generic generators (npc is specfic to sky130).
	Because it is easy to add idenpedently, 
	we implement this as a function wrapper to correctly lay npc
	returns the modified component"""
	# extract licon polygons which are over poly (using booleans)
	licon_comp = comp.extract(layers=[(66,44)])
	poly_comp = comp.extract(layers=[(66,20)])
	existing_npc = comp.extract(layers=[(95,20)])
	# TODO: see about an implemtation using gdsfactory component metadata
	if len(licon_comp.get_polygons()) < 2 and len(poly_comp.get_polygons()) < 2:
		return comp
	liconANDpoly = boolean(licon_comp, poly_comp, layer=(1,2), operation="and")
	if len(existing_npc.get_polygons()) > 1:
		liconANDpoly = boolean(liconANDpoly, existing_npc, layer=(1,2), operation="A-B")
	licon_polygons = liconANDpoly.get_polygons(as_array=False)
	# iterate through all licon and create npc (ignore merges for now)
	npc_polygons = list()
	for licon_polygon in licon_polygons:
		bbox = licon_polygon.bounding_box()
		licon_polygonxmin = bbox[0][0]
		licon_polygonymin = bbox[0][1]
		licon_polygonxmax = bbox[1][0]
		licon_polygonymax = bbox[1][1]
		padding_points = [
			[licon_polygonxmin - 0.1, licon_polygonymin - 0.1],
			[licon_polygonxmax + 0.1, licon_polygonymin - 0.1],
			[licon_polygonxmax + 0.1, licon_polygonymax + 0.1],
			[licon_polygonxmin - 0.1, licon_polygonymax + 0.1],
		]
		npc_polygons.append(Polygon(padding_points, layer=(95,20)))
	# determine which npc polygons should be merged 
	# also merge them by adding a polygon over them 
	# naive approach, n^2 complexity
	npc_merged_polygons = list()
	for i, npc_polygon in enumerate(npc_polygons):
		for j, other_polygon in enumerate(npc_polygons):
			# use the fact that all npc polys have the same width (at this point)
			yviolation = abs(npc_polygon.center[1] - other_polygon.center[1]) < 0.64#0.27+0.37
			xviolation = abs(npc_polygon.center[0] - other_polygon.center[0]) < 0.64
			if i==j:#skip same polygon
				continue
			elif (xviolation and yviolation):
				nxmax = max(npc_polygon.xmax, other_polygon.xmax)
				nxmin = min(npc_polygon.xmin, other_polygon.xmin)
				nymax = max(npc_polygon.ymax, other_polygon.ymax)
				nymin = min(npc_polygon.ymin, other_polygon.ymin)
				points = [
					[nxmin,nymin],
					[nxmax,nymin],
					[nxmax,nymax],
					[nxmin,nymax],
				]
				npc_merged_polygons.append(Polygon(points=points,layer=(95,20)))
	# add npc and return
	npc_polygons_to_add = npc_polygons + npc_merged_polygons
	comp.add(npc_polygons_to_add)
	return comp



from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.rectangle import rectangle
from glayout.pdk.mappedpdk import MappedPDK
from typing import Optional
from glayout.primitives.via_gen import via_array
from glayout.pdk.util.comp_utils import prec_array, to_decimal, to_float
from glayout.pdk.util.port_utils import rename_ports_by_orientation, add_ports_perimeter, print_ports
from pydantic import validate_arguments
from glayout.routing.straight_route import straight_route
from decimal import ROUND_UP, Decimal
from glayout.spice import Netlist

@validate_arguments
def __get_mimcap_layerconstruction_info(pdk: MappedPDK) -> tuple[str,str]:
	"""returns the glayer metal below and glayer metal above capmet
	args: pdk
	"""
	capmettop = pdk.layer_to_glayer(pdk.get_grule("capmet")["capmettop"])
	capmetbottom = pdk.layer_to_glayer(pdk.get_grule("capmet")["capmetbottom"])
	pdk.has_required_glayers(["capmet",capmettop,capmetbottom])
	pdk.activate()
	return capmettop, capmetbottom

def __generate_mimcap_netlist(pdk: MappedPDK, size: tuple[float, float]) -> Netlist:
	return Netlist(
		circuit_name="MIMCap",
		nodes = ['V1', 'V2'],
		source_netlist=""".subckt {circuit_name} {nodes} l=1 w=1
X1 V1 V2 {model} l={{l}} w={{w}}
.ends {circuit_name}""",
		instance_format="X{name} {nodes} {circuit_name} l={length} w={width}",
		parameters={
			'model': pdk.models['mimcap'],
			'length': size[0],
			'width': size[1]
		}
	)

def __generate_mimcap_array_netlist(mimcap_netlist: Netlist, num_caps: int) -> Netlist:
	arr_netlist = Netlist(
		circuit_name="MIMCAP_ARR",
		nodes = ['V1', 'V2']
	)

	for _ in range(num_caps):
		arr_netlist.connect_netlist(
			mimcap_netlist,
			[]
		)

	return arr_netlist

@cell
def mimcap(
    pdk: MappedPDK, size: tuple[float,float]=(5.0, 5.0)
) -> Component:
    """create a mimcap
    args:
    pdk=pdk to use
    size=tuple(float,float) size of cap
    ****Note: size is the size of the capmet layer
    ports:
    top_met_...all edges, this is the metal over the capmet
    bottom_met_...all edges, this is the metal below capmet
    """
    size = pdk.snap_to_2xgrid(size)
    # error checking and
    capmettop, capmetbottom = __get_mimcap_layerconstruction_info(pdk)
    # create top component
    mim_cap = Component()
    mim_cap << rectangle(size=size, layer=pdk.get_glayer("capmet"), centered=True)
    top_met_ref = mim_cap << via_array(
        pdk, capmetbottom, capmettop, size=size, minus1=True, lay_bottom=False
    )
    bottom_met_enclosure = pdk.get_grule(capmetbottom,"capmet")["min_enclosure"]
    mim_cap.add_padding(layers=(pdk.get_glayer(capmetbottom),),default=bottom_met_enclosure)
    # flatten and create ports
    mim_cap = add_ports_perimeter(mim_cap, layer=pdk.get_glayer(capmetbottom), prefix="bottom_met_")
    mim_cap.add_ports(top_met_ref.get_ports_list())

    component = rename_ports_by_orientation(mim_cap).flatten()

    # netlist generation
    component.info['netlist'] = __generate_mimcap_netlist(pdk, size)

    return component

@cell
def mimcap_array(pdk: MappedPDK, rows: int, columns: int, size: tuple[float,float] = (5.0,5.0), rmult: Optional[int]=1) -> Component:
	"""create mimcap array
	args:
	pdk to use
	size = tuple(float,float) size of a single cap
	****Note: size is the size of the capmet layer
	ports:
	cap_x_y_top_met_...all edges, this is the metal over the capmet in row x, col y
	cap_x_y_bottom_met_...all edges, this is the metal below capmet in row x, col y
	"""
	capmettop, capmetbottom = __get_mimcap_layerconstruction_info(pdk)
	mimcap_arr = Component()
	# create the mimcap array
	mimcap_single = mimcap(pdk, size)
	mimcap_space = pdk.get_grule("capmet")["min_separation"] #+ evaluate_bbox(mimcap_single)[0]
	array_ref = mimcap_arr << prec_array(mimcap_single, rows, columns, spacing=2*[mimcap_space])
	mimcap_arr.add_ports(array_ref.get_ports_list())
	# create a list of ports that should be routed to connect the array
	port_pairs = list()
	for rownum in range(rows):
		for colnum in range(columns):
			bl_mimcap = f"row{rownum}_col{colnum}_"
			right_mimcap = f"row{rownum}_col{colnum+1}_"
			top_mimcap = f"row{rownum+1}_col{colnum}_"
			for level,layer in [("bottom_met_",capmetbottom),("top_met_",capmettop)]:
				bl_east_port = mimcap_arr.ports.get(bl_mimcap+level+"E")
				r_west_port = mimcap_arr.ports.get(right_mimcap+level+"W")
				bl_north_port = mimcap_arr.ports.get(bl_mimcap+level+"N")
				top_south_port = mimcap_arr.ports.get(top_mimcap+level+"S")
				if rownum == rows-1 and colnum == columns-1:
					continue
				elif rownum == rows-1:
					port_pairs.append((bl_east_port,r_west_port,layer))
				elif colnum == columns-1:
					port_pairs.append((bl_north_port,top_south_port,layer))
				else:
					port_pairs.append((bl_east_port,r_west_port,layer))
					port_pairs.append((bl_north_port,top_south_port,layer))
	for port_pair in port_pairs:
		mimcap_arr << straight_route(pdk,port_pair[0],port_pair[1],width=rmult*pdk.get_grule(port_pair[2])["min_width"])

	# add netlist
	mimcap_arr.info['netlist'] = __generate_mimcap_array_netlist(mimcap_single.info['netlist'], rows * columns)

	return mimcap_arr.flatten()





from glayout.pdk.mappedpdk import MappedPDK
from pydantic import validate_arguments
from gdsfactory.component import Component
from glayout.primitives.fet import nmos, pmos, multiplier
from glayout.pdk.util.comp_utils import evaluate_bbox
from typing import Literal, Union
from glayout.pdk.util.port_utils import rename_ports_by_orientation, rename_ports_by_list
from glayout.pdk.util.comp_utils import prec_ref_center
from glayout.routing.straight_route import straight_route
from gdsfactory.functions import transformed
from glayout.primitives.guardring import tapring
from glayout.pdk.util.port_utils import add_ports_perimeter
from gdsfactory.cell import clear_cache


#from glayout.placement.two_transistor_interdigitized import two_nfet_interdigitized; from glayout.pdk.sky130_mapped import sky130_mapped_pdk as pdk; biasParams=[6,2,4]; rmult=2

@validate_arguments
def two_transistor_interdigitized(
    pdk: MappedPDK,
    numcols: int,
    deviceA_and_B: Literal["nfet", "pfet"],
    dummy: Union[bool, tuple[bool, bool]] = True,
    **kwargs
) -> Component:
    """place two transistors in a single row with interdigitized placement
    Currently only supports two of the same transistor (same devices)
    Place follows an ABABAB... pattern
    args:
    pdk = MappedPDK to use
    numcols = a single col is actually one col for both transistors (so AB). 2 cols = ABAB ... so on
    deviceA_and_B = the device to place for both transistors (either nfet or pfet)
    dummy = place dummy at the edges of the interdigitized place (true by default). you can specify tuple to place only on one side
    kwargs = key word arguments for device. 
    ****NOTE: These are the same as glayout.primitives.fet.multiplier arguments EXCLUDING dummy, sd_route_extension, and pdk options
    """
    if isinstance(dummy, bool):
        dummy = (dummy, dummy)
    # override kwargs for needed options
    kwargs["sd_route_extension"] = 0
    kwargs["gate_route_extension"] = 0
    kwargs["sdlayer"] = "n+s/d" if deviceA_and_B == "nfet" else "p+s/d"
    kwargs["pdk"] = pdk
    # create devices dummy l/r and A/B (change extension options)
    kwargs["dummy"] = (True,False) if dummy[0] else False
    lefttmost_devA = multiplier(**kwargs)
    kwargs["dummy"] = False
    center_devA = multiplier(**kwargs)
    devB_sd_extension = pdk.util_max_metal_seperation() + abs(center_devA.ports["drain_N"].center[1]-center_devA.ports["diff_N"].center[1])
    devB_gate_extension = pdk.util_max_metal_seperation() + abs(center_devA.ports["row0_col0_gate_S"].center[1]-center_devA.ports["gate_S"].center[1])
    kwargs["sd_route_extension"] = pdk.snap_to_2xgrid(devB_sd_extension)
    kwargs["gate_route_extension"] = pdk.snap_to_2xgrid(devB_gate_extension)
    center_devB = multiplier(**kwargs)
    kwargs["dummy"] = (False,True) if dummy[1] else False
    rightmost_devB = multiplier(**kwargs)
    # place devices
    idplace = Component()
    dims = evaluate_bbox(center_devA)
    xdisp = pdk.snap_to_2xgrid(dims[0]+pdk.get_grule("active_diff")["min_separation"])
    refs = list()
    for i in range(2*numcols):
        if i==0:
            refs.append(idplace << lefttmost_devA)
        elif i==((2*numcols)-1):
            refs.append(idplace << rightmost_devB)
        elif i%2: # is odd (so device B)
            refs.append(idplace << center_devB)
        else: # not i%2 == i is even (so device A)
            refs.append(idplace << center_devA)
        refs[-1].movex(i*(xdisp))
        devletter = "B" if i%2 else "A"
        prefix=devletter+"_"+str(int(i/2))+"_"
        idplace.add_ports(refs[-1].get_ports_list(), prefix=prefix)
    #issue somehwere before line 89
    # merge s/d layer for all transistors
    idplace << straight_route(pdk, refs[0].ports["plusdoped_W"],refs[-1].ports["plusdoped_E"])
    # create s/d/gate connections extending over entire row
    A_src = idplace << rename_ports_by_orientation(rename_ports_by_list(straight_route(pdk, refs[0].ports["source_W"], refs[-1].ports["source_E"]), [("route_","_")]))
    B_src = idplace << rename_ports_by_orientation(rename_ports_by_list(straight_route(pdk, refs[-1].ports["source_E"], refs[0].ports["source_W"]), [("route_","_")]))
    A_drain = idplace << rename_ports_by_orientation(rename_ports_by_list(straight_route(pdk, refs[0].ports["drain_W"], refs[-1].ports["drain_E"]), [("route_","_")]))
    B_drain = idplace << rename_ports_by_orientation(rename_ports_by_list(straight_route(pdk, refs[-1].ports["drain_E"], refs[0].ports["drain_W"]), [("route_","_")]))
    A_gate = idplace << rename_ports_by_orientation(rename_ports_by_list(straight_route(pdk, refs[0].ports["gate_W"], refs[-1].ports["gate_E"]), [("route_","_")]))
    B_gate = idplace << rename_ports_by_orientation(rename_ports_by_list(straight_route(pdk, refs[-1].ports["gate_E"], refs[0].ports["gate_W"]), [("route_","_")]))
    # add route ports and return
    prefixes = ["A_source","B_source","A_drain","B_drain","A_gate","B_gate"]
    for i, ref in enumerate([A_src, B_src, A_drain, B_drain, A_gate, B_gate]):
        idplace.add_ports(ref.get_ports_list(),prefix=prefixes[i])
    idplace = transformed(prec_ref_center(idplace))
    idplace.unlock()
    return idplace


@validate_arguments
def two_nfet_interdigitized(
    pdk: MappedPDK,
    numcols: int,
    dummy: Union[bool, tuple[bool, bool]] = True,
    with_substrate_tap: bool = True,
    with_tie: bool = True,
    tie_layers: tuple[str,str]=("met2","met1"),
    **kwargs
) -> Component:
    """Currently only supports two of the same nfet instances. does NOT support multipliers (currently)
    Place follows an ABABAB... pattern
    args:
    pdk = MappedPDK to use
    numcols = a single col is actually one col for both nfets (so AB). 2 cols = ABAB ... so on
    dummy = place dummy at the edges of the interdigitized place (true by default). you can specify tuple to place only on one side
    kwargs = key word arguments for multiplier. 
    ****NOTE: These are the same as glayout.primitives.fet.multiplier arguments EXCLUDING dummy, sd_route_extension, and pdk options
    tie_layers: tuple[str,str] specifying (horizontal glayer, vertical glayer) or well tie ring. default=("met2","met1")
    """
    base_multiplier = two_transistor_interdigitized(pdk, numcols, "nfet", dummy, **kwargs)
    # tie
    if with_tie:
        tap_separation = max(
            pdk.util_max_metal_seperation(),
            pdk.get_grule("active_diff", "active_tap")["min_separation"],
        )
        tap_separation += pdk.get_grule("p+s/d", "active_tap")["min_enclosure"]
        tap_encloses = (
            2 * (tap_separation + base_multiplier.xmax),
            2 * (tap_separation + base_multiplier.ymax),
        )
        tiering_ref = base_multiplier << tapring(
            pdk,
            enclosed_rectangle=tap_encloses,
            sdlayer="p+s/d",
            horizontal_glayer=tie_layers[0],
            vertical_glayer=tie_layers[1],
        )
        base_multiplier.add_ports(tiering_ref.get_ports_list(), prefix="welltie_")
        try:
            base_multiplier<<straight_route(pdk,base_multiplier.ports["A_0_dummy_L_gsdcon_top_met_W"],base_multiplier.ports["welltie_W_top_met_W"],glayer2="met1")
        except KeyError:
            pass
        try:
            base_multiplier<<straight_route(pdk,base_multiplier.ports[f"B_{numcols-1}_dummy_R_gsdcon_top_met_E"],base_multiplier.ports["welltie_E_top_met_E"],glayer2="met1")
        except KeyError:
            pass
    # add pwell
    base_multiplier.add_padding(
        layers=(pdk.get_glayer("pwell"),),
        default=pdk.get_grule("pwell", "active_tap")["min_enclosure"],
    )
    # add substrate tap
    base_multiplier = add_ports_perimeter(base_multiplier,layer=pdk.get_glayer("pwell"),prefix="well_")
    # add substrate tap if with_substrate_tap
    if with_substrate_tap:
        substrate_tap_separation = pdk.get_grule("dnwell", "active_tap")[
            "min_separation"
        ]
        substrate_tap_encloses = (
            2 * (substrate_tap_separation + base_multiplier.xmax),
            2 * (substrate_tap_separation + base_multiplier.ymax),
        )
        ringtoadd = tapring(
            pdk,
            enclosed_rectangle=substrate_tap_encloses,
            sdlayer="p+s/d",
            horizontal_glayer="met2",
            vertical_glayer="met1",
        )
        tapring_ref = base_multiplier << ringtoadd
        base_multiplier.add_ports(tapring_ref.get_ports_list(),prefix="substratetap_")
    return base_multiplier



from gdsfactory.cell import cell, clear_cache
from gdsfactory.component import Component, copy
from gdsfactory.component_reference import ComponentReference
from gdsfactory.components.rectangle import rectangle
from glayout.pdk.mappedpdk import MappedPDK
from typing import Optional, Union
from glayout.primitives.fet import nmos, pmos, multiplier
from glayout.components.diff_pair import diff_pair
from glayout.primitives.guardring import tapring
from glayout.primitives.mimcap import mimcap_array, mimcap
from glayout.routing.L_route import L_route
from glayout.routing.c_route import c_route
from glayout.primitives.via_gen import via_stack, via_array
from gdsfactory.routing.route_quad import route_quad
from glayout.pdk.util.comp_utils import evaluate_bbox, prec_ref_center, movex, movey, to_decimal, to_float, move, align_comp_to_port, get_padding_points_cc
from glayout.pdk.util.port_utils import rename_ports_by_orientation, rename_ports_by_list, add_ports_perimeter, print_ports, set_port_orientation, rename_component_ports
from glayout.routing.straight_route import straight_route
from glayout.pdk.util.snap_to_grid import component_snap_to_grid
from pydantic import validate_arguments
from glayout.placement.two_transistor_interdigitized import two_nfet_interdigitized
from glayout.spice import Netlist



@validate_arguments
def __create_sharedgatecomps(pdk: MappedPDK, rmult: int, half_pload: tuple[float,float,int]) -> tuple:
    # add diffpair current mirror loads (this is a pmos current mirror split into 2 for better matching/compensation)
    shared_gate_comps = Component("shared gate components")
    # create the 2*2 multiplier transistors (placed twice later)
    twomultpcomps = Component("2 multiplier shared gate comps")
    pcompR = multiplier(pdk, "p+s/d", width=half_pload[0], length=half_pload[1], fingers=half_pload[2], dummy=True,rmult=rmult).copy()
    tapref = pcompR << tapring(pdk, evaluate_bbox(pcompR,padding=0.3+pdk.get_grule("n+s/d", "active_tap")["min_enclosure"]),"n+s/d","met1","met1")
    pcompR.add_padding(layers=(pdk.get_glayer("nwell"),), default=pdk.get_grule("active_tap", "nwell")["min_enclosure"])
    pcompR.add_ports(tapref.get_ports_list(),prefix="welltap_")
    pcompR << straight_route(pdk,pcompR.ports["dummy_L_gsdcon_top_met_W"],pcompR.ports["welltap_W_top_met_W"],glayer2="met1")
    pcompR << straight_route(pdk,pcompR.ports["dummy_R_gsdcon_top_met_W"],pcompR.ports["welltap_E_top_met_E"],glayer2="met1")
    pcompL = pcompR.copy()
    pcomp_AB_spacing = max(2*pdk.util_max_metal_seperation() + 6*pdk.get_grule("met4")["min_width"],pdk.get_grule("p+s/d")["min_separation"])
    _prefL = (twomultpcomps << pcompL).movex(-1 * pcompL.xmax - pcomp_AB_spacing/2)
    _prefR = (twomultpcomps << pcompR).movex(-1 * pcompR.xmin + pcomp_AB_spacing/2)
    twomultpcomps.add_ports(_prefL.get_ports_list(),prefix="L_")
    twomultpcomps.add_ports(_prefR.get_ports_list(),prefix="R_")
    twomultpcomps << route_quad(_prefL.ports["gate_W"], _prefR.ports["gate_E"], layer=pdk.get_glayer("met2"))
    # center
    relative_dim_comp = multiplier(
        pdk, "p+s/d", width=half_pload[0], length=half_pload[1], fingers=4, dummy=False, rmult=rmult
    )
    # TODO: figure out single dim spacing rule then delete both test delete and this
    single_dim = to_decimal(relative_dim_comp.xmax) + to_decimal(0.11) + to_decimal(half_pload[1])/2
    LRplusdopedPorts = list()
    LRgatePorts = list()
    LRdrainsPorts = list()
    LRsourcesPorts = list()
    LRdummyports = list()
    for i in [-2, -1, 1, 2]:
        dummy = False
        appenddummy = None
        extra_t = 0
        if i == -2:
            dummy = [True, False]
            appenddummy="L"
            pcenterfourunits = multiplier(
                pdk, "p+s/d", width=half_pload[0], length=half_pload[1], fingers=4, dummy=dummy, rmult=rmult
            )
            extra_t = -1 * single_dim
        elif i == 2:
            dummy = [False, True]
            appenddummy="R"
            pcenterfourunits = multiplier(
                pdk, "p+s/d", width=half_pload[0], length=half_pload[1], fingers=4, dummy=dummy, rmult=rmult
            )
            extra_t = single_dim
        else:
            pcenterfourunits = relative_dim_comp
        pref_ = prec_ref_center(pcenterfourunits).movex(pdk.snap_to_2xgrid(to_float(i * single_dim + extra_t)))
        shared_gate_comps.add(pref_)
        if appenddummy:
            LRdummyports+= [pref_.ports["dummy_"+appenddummy+"_gsdcon_top_met_N"]]
        LRplusdopedPorts += [pref_.ports["plusdoped_W"] , pref_.ports["plusdoped_E"]]
        LRgatePorts += [pref_.ports["gate_W"],pref_.ports["gate_E"]]
        LRdrainsPorts += [pref_.ports["source_W"],pref_.ports["source_E"]]
        LRsourcesPorts += [pref_.ports["drain_W"],pref_.ports["drain_E"]]
    # combine the two multiplier top and bottom with the 4 multiplier center row
    ytranslation_pcenter = 2 * pcenterfourunits.ymax + 5*pdk.util_max_metal_seperation()
    ptop_AB = (shared_gate_comps << twomultpcomps).movey(ytranslation_pcenter)
    pbottom_AB = (shared_gate_comps << twomultpcomps).movey(-1 * ytranslation_pcenter)

    return shared_gate_comps, ptop_AB, pbottom_AB, LRplusdopedPorts, LRgatePorts, LRdrainsPorts, LRsourcesPorts, LRdummyports



def __route_sharedgatecomps(pdk: MappedPDK, shared_gate_comps, via_location, ptop_AB, pbottom_AB, LRplusdopedPorts, LRgatePorts, LRdrainsPorts, LRsourcesPorts,LRdummyports) -> Component:
    _max_metal_seperation_ps = pdk.util_max_metal_seperation()
    # ground dummy transistors of the 4 center multipliers
    shared_gate_comps << straight_route(pdk,LRdummyports[0],pbottom_AB.ports["L_welltap_N_top_met_S"],glayer2="met1")
    shared_gate_comps << straight_route(pdk,LRdummyports[1],pbottom_AB.ports["R_welltap_N_top_met_S"],glayer2="met1")
    # connect p+s/d layer of the transistors
    shared_gate_comps << route_quad(LRplusdopedPorts[0],LRplusdopedPorts[-1],layer=pdk.get_glayer("p+s/d"))
    # connect drain of the left 2 and right 2, short sources of all 4
    shared_gate_comps << route_quad(LRdrainsPorts[0],LRdrainsPorts[3],layer=LRdrainsPorts[0].layer)
    shared_gate_comps << route_quad(LRdrainsPorts[4],LRdrainsPorts[7],layer=LRdrainsPorts[0].layer)
    shared_gate_comps << route_quad(LRsourcesPorts[0],LRsourcesPorts[-1],layer=LRsourcesPorts[0].layer)
    pcomps_2L_2R_sourcevia = shared_gate_comps << via_stack(pdk,pdk.layer_to_glayer(LRsourcesPorts[0].layer), "met4")
    pcomps_2L_2R_sourcevia.movey(evaluate_bbox(pcomps_2L_2R_sourcevia.parent.extract(layers=[LRsourcesPorts[0].layer,]))[1]/2 + LRsourcesPorts[0].center[1])
    shared_gate_comps.add_ports(pcomps_2L_2R_sourcevia.get_ports_list(),prefix="2L2Rsrcvia_")
    # short all the gates
    shared_gate_comps << route_quad(LRgatePorts[0],LRgatePorts[-1],layer=pdk.get_glayer("met2"))
    shared_gate_comps.add_ports(ptop_AB.get_ports_list(),prefix="ptopAB_")
    shared_gate_comps.add_ports(pbottom_AB.get_ports_list(),prefix="pbottomAB_")
    # short all gates of shared_gate_comps
    pcenter_gate_route_extension = pdk.snap_to_2xgrid(shared_gate_comps.xmax - min(ptop_AB.ports["R_gate_E"].center[0], LRgatePorts[-1].center[0]) - pdk.get_grule("active_diff")["min_width"])
    pcenter_l_croute = shared_gate_comps << c_route(pdk, ptop_AB.ports["L_gate_W"], pbottom_AB.ports["L_gate_W"],extension=pcenter_gate_route_extension)
    pcenter_r_croute = shared_gate_comps << c_route(pdk, ptop_AB.ports["R_gate_E"], pbottom_AB.ports["R_gate_E"],extension=pcenter_gate_route_extension)
    shared_gate_comps << straight_route(pdk, LRgatePorts[0], pcenter_l_croute.ports["con_N"])
    shared_gate_comps << straight_route(pdk, LRgatePorts[-1], pcenter_r_croute.ports["con_N"])
    # connect drain of A to the shorted gates
    shared_gate_comps << L_route(pdk,ptop_AB.ports["L_source_W"],pcenter_l_croute.ports["con_N"])
    shared_gate_comps << straight_route(pdk,pbottom_AB.ports["R_source_E"],pcenter_r_croute.ports["con_N"])
    # connect source of A to the drain of 2L
    pcomps_route_A_drain_extension = shared_gate_comps.xmax-max(ptop_AB.ports["R_drain_E"].center[0], LRdrainsPorts[-1].center[0])+_max_metal_seperation_ps
    pcomps_route_A_drain = shared_gate_comps << c_route(pdk, ptop_AB.ports["L_drain_W"], LRdrainsPorts[0], extension=pcomps_route_A_drain_extension)
    row_rectangle_routing = rectangle(layer=ptop_AB.ports["L_drain_W"].layer,size=(pbottom_AB.ports["R_source_N"].width,pbottom_AB.ports["R_source_W"].width)).copy()
    Aextra_top_connection = align_comp_to_port(row_rectangle_routing, pbottom_AB.ports["R_source_N"], ('c','t')).movey(row_rectangle_routing.ymax + _max_metal_seperation_ps)
    shared_gate_comps.add(Aextra_top_connection)
    shared_gate_comps << straight_route(pdk,Aextra_top_connection.ports["e4"],pbottom_AB.ports["R_drain_N"])
    shared_gate_comps << L_route(pdk,pcomps_route_A_drain.ports["con_S"], Aextra_top_connection.ports["e1"],viaoffset=(False,True))
    # connect source of B to drain of 2R
    pcomps_route_B_source_extension = shared_gate_comps.xmax-max(LRsourcesPorts[-1].center[0],ptop_AB.ports["R_source_E"].center[0])+_max_metal_seperation_ps
    mimcap_connection_ref = shared_gate_comps << c_route(pdk, ptop_AB.ports["R_source_E"], LRdrainsPorts[-1],extension=pcomps_route_B_source_extension,viaoffset=(True,False))
    bottom_pcompB_floating_port = set_port_orientation(movey(movex(pbottom_AB.ports["L_source_E"].copy(),5*_max_metal_seperation_ps), destination=Aextra_top_connection.ports["e1"].center[1]+Aextra_top_connection.ports["e1"].width+_max_metal_seperation_ps),"S")
    pmos_bsource_2Rdrain_v = shared_gate_comps << L_route(pdk,pbottom_AB.ports["L_source_E"],bottom_pcompB_floating_port,vglayer="met3")
    # fix the extension when the top row of transistors extends farther than the middle row
    if LRdrainsPorts[-1].center[0] < ptop_AB.ports["R_source_E"].center[0]:
        pcomps_route_B_source_extension += ptop_AB.ports["R_source_E"].center[0] - LRdrainsPorts[-1].center[0]
    shared_gate_comps << c_route(pdk, LRdrainsPorts[-1], set_port_orientation(bottom_pcompB_floating_port,"E"),extension=pcomps_route_B_source_extension,viaoffset=(True,False))
    pmos_bsource_2Rdrain_v_center = via_stack(pdk,"met2","met3",fulltop=True)
    shared_gate_comps.add(align_comp_to_port(pmos_bsource_2Rdrain_v_center, bottom_pcompB_floating_port,('r','t')))
    # connect drain of B to each other directly over where the diffpair top left drain will be
    pmos_bdrain_diffpair_v = shared_gate_comps << via_stack(pdk, "met2","met5",fullbottom=True)
    pmos_bdrain_diffpair_v = align_comp_to_port(pmos_bdrain_diffpair_v, movex(pbottom_AB.ports["L_gate_S"].copy(),destination=via_location))
    pmos_bdrain_diffpair_v.movey(0-_max_metal_seperation_ps)
    pcomps_route_B_drain_extension = shared_gate_comps.xmax-ptop_AB.ports["R_drain_E"].center[0]+_max_metal_seperation_ps
    shared_gate_comps << c_route(pdk, ptop_AB.ports["R_drain_E"], pmos_bdrain_diffpair_v.ports["bottom_met_E"],extension=pcomps_route_B_drain_extension +_max_metal_seperation_ps)
    shared_gate_comps << c_route(pdk, pbottom_AB.ports["L_drain_W"], pmos_bdrain_diffpair_v.ports["bottom_met_W"],extension=pcomps_route_B_drain_extension +_max_metal_seperation_ps)
    shared_gate_comps.add_ports(pmos_bdrain_diffpair_v.get_ports_list(),prefix="minusvia_")
    shared_gate_comps.add_ports(mimcap_connection_ref.get_ports_list(),prefix="mimcap_connection_")
    return shared_gate_comps

def differential_to_single_ended_converter_netlist(pdk: MappedPDK, half_pload: tuple[float, float, int]) -> Netlist:
    return Netlist(
        circuit_name="DIFF_TO_SINGLE",
        nodes=['VIN', 'VOUT', 'VSS', 'VSS2'],
        source_netlist=""".subckt {circuit_name} {nodes} l=1 w=1 mt=1 mb=1
XTOP1 V1   VIN VSS  VSS {model} l={{l}} w={{w}} m={{mt}}
XTOP2 VSS2 VIN VSS  VSS {model} l={{l}} w={{w}} m={{mt}}
XBOT1 VIN  VIN V1   VSS {model} l={{l}} w={{w}} m={{mb}}
XBOT2 VOUT VIN VSS2 VSS {model} l={{l}} w={{w}} m={{mb}}
.ends {circuit_name}""",
        instance_format="X{name} {nodes} {circuit_name} l={length} w={width} mt={mult_top} mb={mult_bot}",
        parameters={
            'model': pdk.models['pfet'],
            'width': half_pload[0],
            'length': half_pload[1],
            'mult_top': 4 * 2,
            'mult_bot': half_pload[2] * 2
        }
    )

def differential_to_single_ended_converter(pdk: MappedPDK, rmult: int, half_pload: tuple[float,float,int], via_xlocation) -> tuple:
    clear_cache()
    pmos_comps, ptop_AB, pbottom_AB, LRplusdopedPorts, LRgatePorts, LRdrainsPorts, LRsourcesPorts, LRdummyports = __create_sharedgatecomps(pdk, rmult,half_pload)
    clear_cache()
    pmos_comps = __route_sharedgatecomps(pdk, pmos_comps, via_xlocation, ptop_AB, pbottom_AB, LRplusdopedPorts, LRgatePorts, LRdrainsPorts, LRsourcesPorts, LRdummyports)

    pmos_comps.info['netlist'] = differential_to_single_ended_converter_netlist(pdk, half_pload)

    return pmos_comps



from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.port import Port
from glayout.pdk.mappedpdk import MappedPDK
from typing import Optional, Union
from math import isclose
from glayout.primitives.via_gen import via_stack
from gdsfactory.routing.route_quad import route_quad
from gdsfactory.components.rectangle import rectangle
from glayout.pdk.util.comp_utils import evaluate_bbox, get_primitive_rectangle
from glayout.pdk.util.port_utils import add_ports_perimeter, rename_ports_by_orientation, rename_ports_by_list, print_ports, set_port_width, set_port_orientation, get_orientation
from pydantic import validate_arguments
from gdsfactory.snap import snap_to_grid


@validate_arguments
def __fill_empty_viastack__macro(pdk: MappedPDK, glayer: str, size: tuple[float,float]) -> Component:
	"""returns a rectangle with ports that pretend to be viastack ports"""
	comp = rectangle(size=size,layer=pdk.get_glayer(glayer),centered=True)
	return rename_ports_by_orientation(rename_ports_by_list(comp,replace_list=[("e","top_met_")])).flatten()

@cell
def c_route(
	pdk: MappedPDK, 
	edge1: Port, 
	edge2: Port, 
	extension: Optional[float]=0.5, 
	width1: Optional[float] = None, 
	width2: Optional[float] = None,
	cwidth: Optional[float] = None,
	e1glayer: Optional[str] = None, 
	e2glayer: Optional[str] = None, 
	cglayer: Optional[str] = None, 
	viaoffset: Optional[Union[bool,tuple[Optional[bool],Optional[bool]]]]=(True,True),
	fullbottom: Optional[bool] = False
) -> Component:
	"""creates a C shaped route between two Ports.
	
	edge1--|
	       |
	edge2--|
	
	REQUIRES: ports be parralel vertical or horizontal edges
	****NOTE: does no drc error checking (creates a dumb route)
	args:
	pdk = pdk to use
	edge1 = first port
	edge2 = second port
	width1 = optional will default to edge1 width if None
	width2 = optional will default to edge2 width if None
	e1glayer = glayer for the parts connecting to the edge1. Default to layer of edge1
	e2glayer = glayer for the parts connecting to the edge2. Default to layer of edge2
	cglayer = glayer for the connection part (part that goes through a via) defaults to e1glayer met+1
	viaoffset = offsets the via so that it is flush with the cglayer (may be needed for drc) i.e. -| vs _|
	- True offsets via towards the other via
	- False offsets via away from the other via
	- None means center (no offset)
	****NOTE: viaoffset pushes both vias towards each other slightly
	"""
	# error checking and figure out args
	if round(edge1.orientation) % 90 or round(edge2.orientation) % 90:
		raise ValueError("Ports must be vertical or horizontal")
	if not isclose(edge1.orientation,edge2.orientation):
		raise ValueError("Ports must be parralel and have same orientation")
	width1 = width1 if width1 else edge1.width
	width2 = width2 if width2 else edge1.width
	e1glayer = e1glayer if e1glayer else pdk.layer_to_glayer(edge1.layer)
	e2glayer = e2glayer if e2glayer else pdk.layer_to_glayer(edge2.layer)
	eglayer_plusone = "met" + str(int(e1glayer[-1])+1)
	cglayer = cglayer if cglayer else eglayer_plusone
	if not "met" in e1glayer or not "met" in e2glayer or not "met" in cglayer:
		raise ValueError("given layers must be metals")
	viaoffset = (None, None) if viaoffset is None else viaoffset
	if isinstance(viaoffset,bool):
		viaoffset = (True,True) if viaoffset else (False,False)
	pdk.has_required_glayers([e1glayer,e2glayer,cglayer])
	pdk.activate()
	extension = snap_to_grid(extension)
	# create route
	croute = Component()
	viastack1 = via_stack(pdk,e1glayer,cglayer,fullbottom=fullbottom,assume_bottom_via=True)
	viastack2 = via_stack(pdk,e2glayer,cglayer,fullbottom=fullbottom,assume_bottom_via=True)
	if e1glayer == e2glayer:
		pass
	elif e1glayer == cglayer:
		viastack1 = __fill_empty_viastack__macro(pdk,e1glayer,size=evaluate_bbox(viastack2))
	elif e2glayer == cglayer:
		viastack2 = __fill_empty_viastack__macro(pdk,e2glayer,size=evaluate_bbox(viastack1))
	# find extension
	e1_length = snap_to_grid(extension + evaluate_bbox(viastack1)[0])
	e2_length = snap_to_grid(extension + evaluate_bbox(viastack2)[0])
	xdiff = snap_to_grid(abs(edge1.center[0] - edge2.center[0]))
	ydiff = snap_to_grid(abs(edge1.center[1] - edge2.center[1]))
	if not isclose(edge1.center[0],edge2.center[0]):
		if round(edge1.orientation) == 0:# facing east
			if edge1.center[0] > edge2.center[0]:
				e2_length += xdiff
			else:
				e1_length += xdiff
		elif round(edge1.orientation) == 180:# facing west
			if edge1.center[0] < edge2.center[0]:
				e2_length += xdiff
			else:
				e1_length += xdiff
	if not isclose(edge1.center[1],edge2.center[1]):
		if round(edge1.orientation) == 270:# facing south
			if edge1.center[1] < edge2.center[1]:
				e2_length += ydiff
			else:
				e1_length += ydiff
		elif round(edge1.orientation) == 90:#facing north
			if edge1.center[1] > edge2.center[1]:
				e2_length += ydiff
			else:
				e1_length += ydiff
	# move into position
	e1_extension_comp = Component("edge1 extension")
	e2_extension_comp = Component("edge2 extension")
	box_dims = [(e1_length, width1),(e2_length, width2)]
	if round(edge1.orientation) == 90 or round(edge1.orientation) == 270:
		box_dims = [(width1, e1_length),(width2, e2_length)]
	rect_c1 = get_primitive_rectangle(size=box_dims[0], layer=pdk.get_glayer(e1glayer))
	rect_c2 = get_primitive_rectangle(size=box_dims[1], layer=pdk.get_glayer(e2glayer))
	rect_c1 = rename_ports_by_orientation(rename_ports_by_list(rect_c1,[("e","e_")]))
	rect_c2 = rename_ports_by_orientation(rename_ports_by_list(rect_c2,[("e","e_")]))
	# TODO: make sure ports match bbox
	e1_extension = e1_extension_comp << rect_c1
	e2_extension = e2_extension_comp << rect_c2
	e1_extension.move(destination=edge1.center)
	e2_extension.move(destination=edge2.center)
	if round(edge1.orientation) == 0:# facing east
		e1_extension.movey(0-evaluate_bbox(e1_extension)[1]/2)
		e2_extension.movey(0-evaluate_bbox(e2_extension)[1]/2)
	elif round(edge1.orientation) == 180:# facing west
		e1_extension.movex(0-evaluate_bbox(e1_extension)[0])
		e2_extension.movex(0-evaluate_bbox(e2_extension)[0])
		e1_extension.movey(0-evaluate_bbox(e1_extension)[1]/2)
		e2_extension.movey(0-evaluate_bbox(e2_extension)[1]/2)
	elif round(edge1.orientation) == 270:# facing south
		e1_extension.movex(0-evaluate_bbox(e1_extension)[0]/2)
		e2_extension.movex(0-evaluate_bbox(e2_extension)[0]/2)
		e1_extension.movey(0-evaluate_bbox(e1_extension)[1])
		e2_extension.movey(0-evaluate_bbox(e2_extension)[1])
	else:#facing north
		e1_extension.movex(0-evaluate_bbox(e1_extension)[0]/2)
		e2_extension.movex(0-evaluate_bbox(e2_extension)[0]/2)
	# place viastacks
	e1_extension_comp.add_ports(e1_extension.get_ports_list())
	e2_extension_comp.add_ports(e2_extension.get_ports_list())
	me1 = e1_extension_comp << viastack1
	me2 = e2_extension_comp << viastack2
	route_ports = [None,None]
	via_flush = snap_to_grid(abs((width1 - evaluate_bbox(viastack1)[0])/2) if viaoffset else 0)
	via_flush1 = via_flush if viaoffset[0] else 0-via_flush
	via_flush1 = 0 if viaoffset[0] is None else via_flush1
	via_flush2 = via_flush if viaoffset[1] else 0-via_flush
	via_flush2 = 0 if viaoffset[1] is None else via_flush2
	if round(edge1.orientation) == 0:# facing east
		me1.move(destination=e1_extension.ports["e_E"].center)
		me2.move(destination=e2_extension.ports["e_E"].center)
		via_flush *= 1 if me1.ymax > me2.ymax else -1
		me1.movex(0-viastack1.xmax).movey(0-via_flush1)
		me2.movex(0-viastack2.xmax).movey(via_flush2)
		me1, me2 = (me1, me2) if (me1.origin[1] > me2.origin[1]) else (me2, me1)
		route_ports = [me1.ports["top_met_N"],me2.ports["top_met_S"]]
	elif round(edge1.orientation) == 180:# facing west
		me1.move(destination=e1_extension.ports["e_W"].center)
		me2.move(destination=e2_extension.ports["e_W"].center)
		via_flush *= 1 if me1.ymax > me2.ymax else -1
		me1.movex(viastack1.xmax).movey(0-via_flush1)
		me2.movex(viastack2.xmax).movey(via_flush2)
		me1, me2 = (me1, me2) if (me1.origin[1] > me2.origin[1]) else (me2, me1)
		route_ports = [me1.ports["top_met_N"],me2.ports["top_met_S"]]
	elif round(edge1.orientation) == 270:# facing south
		me1.move(destination=e1_extension.ports["e_S"].center)
		me2.move(destination=e2_extension.ports["e_S"].center)
		via_flush *= 1 if me1.xmax > me2.xmax else -1
		me1.movey(viastack1.xmax).movex(0-via_flush1)
		me2.movey(viastack2.xmax).movex(via_flush2)
		me1, me2 = (me1, me2) if (me1.origin[0] > me2.origin[0]) else (me2, me1)
		route_ports = [me1.ports["top_met_E"],me2.ports["top_met_W"]]
	else:#facing north
		me1.move(destination=e1_extension.ports["e_N"].center)
		me2.move(destination=e2_extension.ports["e_N"].center)
		via_flush *= 1 if me1.xmax > me2.xmax else -1
		me1.movey(0-viastack1.xmax).movex(0-via_flush1)
		me2.movey(0-viastack2.xmax).movex(via_flush2)
		me1, me2 = (me1, me2) if (me1.origin[0] > me2.origin[0]) else (me2, me1)
		route_ports = [me1.ports["top_met_E"],me2.ports["top_met_W"]]
	# connect extensions, add ports, return
	croute << e1_extension_comp
	croute << e2_extension_comp
	if cwidth:
		route_ports = [set_port_width(port_,cwidth) for port_ in route_ports]
	route_ports[0].width = route_ports[1].width = max(route_ports[0].width, route_ports[1].width)
	cconnection = croute << route_quad(route_ports[0],route_ports[1],layer=pdk.get_glayer(cglayer))
	for i,port_to_add in enumerate(route_ports):
		orta = get_orientation(port_to_add.orientation)
		route_ports[i] = set_port_orientation(port_to_add, orta)
	croute.add_ports(route_ports,prefix="con_")
	return rename_ports_by_orientation(rename_ports_by_list(croute.flatten(), [("con_","con_")]))




from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.port import Port
from glayout.pdk.mappedpdk import MappedPDK
from typing import Optional, Union
from glayout.primitives.via_gen import via_stack, via_array
from glayout.pdk.util.comp_utils import evaluate_bbox, align_comp_to_port, to_decimal, to_float, prec_ref_center, get_primitive_rectangle
from glayout.pdk.util.port_utils import rename_ports_by_orientation, rename_ports_by_list, print_ports, assert_port_manhattan, assert_ports_perpindicular
from decimal import Decimal


@cell
def L_route(
	pdk: MappedPDK,
	edge1: Port,
	edge2: Port,
	vwidth: Optional[float] = None,
	hwidth: Optional[float] = None,
	hglayer: Optional[str] = None,
	vglayer: Optional[str] = None,
	viaoffset: Optional[Union[tuple[bool,bool],bool]]=True,
	fullbottom: bool = True
) -> Component:
	"""creates a L shaped route between two Ports.
	
	edge1
	  |
	  ------|edge2
	
	REQUIRES: 
	- ports (a.k.a. edges) be vertical or horizontal
	- edges be perpindicular to each other
	
	DOES NOT REQUIRE:
	- correct 180 degree orientation of the port (e.g. a south facing port may result in north facing route)
	
	****NOTE: does no drc error checking (creates a dumb route)
	args:
	pdk = pdk to use
	edge1 = first port
	edge2 = second port
	vwidth = optional will default to vertical edge width if None
	hwidth = optional will default to horizontal edge width if None
	hglayer = glayer for vertical route. Defaults to the layer of the edge oriented N/S
	vglayer = glayer for horizontal route. Defaults to the layer of the edge oriented E/W
	viaoffset = push the via away from both edges so that inside corner aligns with via corner
	****via offset can also be specfied as a tuple(bool,bool): movex? if viaoffset[0] and movey? if viaoffset[1]
	fullbottom = fullbottom option for via
	"""
	# error checking, TODO: validate layers
	assert_port_manhattan([edge1,edge2])
	assert_ports_perpindicular(edge1,edge2)
	pdk.activate()
	Lroute = Component()
	# figure out which port is vertical
	vport = None
	hport = None
	edge1_is_EW = bool(round(edge1.orientation + 90) % 180)
	if edge1_is_EW:
		vport, hport = edge1, edge2
	else:
		hport, vport = edge1, edge2
	# arg setup
	vwidth = to_decimal(vwidth if vwidth else vport.width)
	hwidth = to_decimal(hwidth if hwidth else hport.width)
	hglayer = hglayer if hglayer else pdk.layer_to_glayer(vport.layer)
	vglayer = vglayer if vglayer else pdk.layer_to_glayer(hport.layer)
	if isinstance(viaoffset,bool):
		viaoffset = (True,True) if viaoffset else (False,False)
	# compute required dimensions
	hdim_center = to_decimal(vport.center[0]) - to_decimal(hport.center[0])
	vdim_center = to_decimal(hport.center[1]) - to_decimal(vport.center[1])
	hdim = abs(hdim_center) + hwidth/2
	vdim = abs(vdim_center) + vwidth/2
	# create and place vertical and horizontal connections
	hconnect = get_primitive_rectangle(size=to_float((hdim,vwidth)),layer=pdk.get_glayer(hglayer))
	vconnect = get_primitive_rectangle(size=to_float((hwidth,vdim)),layer=pdk.get_glayer(vglayer))
	#xalign
	valign = ("l","c") if hdim_center > 0 else ("r","c")
	halign = ("c","b") if vdim_center > 0 else ("c","t")
	#yalign
	hconnect_ref = align_comp_to_port(hconnect, vport, valign)
	Lroute.add(hconnect_ref)
	vconnect_ref = align_comp_to_port(vconnect, hport, halign)
	Lroute.add(vconnect_ref)
	# create and place via (decide between via stack and via array)
	hv_via = via_stack(pdk, hglayer, vglayer,fullbottom=fullbottom,fulltop=True)
	hv_via_dims = evaluate_bbox(hv_via,True)
	use_stack = hv_via_dims[0] > hwidth or hv_via_dims[1] > vwidth
	if not use_stack:
		hv_via = via_array(pdk, hglayer, vglayer, size=to_float((hwidth,vwidth)), lay_bottom=True)
	h_to_v_via_ref = prec_ref_center(hv_via)
	Lroute.add(h_to_v_via_ref)
	h_to_v_via_ref.move(destination=(hport.center[0], vport.center[1]))
	if viaoffset[0] or viaoffset[1]:
		viadim_osx = evaluate_bbox(h_to_v_via_ref,True)[0]/2
		viaxofs = abs(hwidth/2-viadim_osx)
		viaxofs = to_float(viaxofs if hdim_center > 0 else -1*viaxofs)
		viaxofs = viaxofs if viaoffset[0] else 0
		viadim_osy = evaluate_bbox(h_to_v_via_ref,True)[1]/2
		viayofs = abs(vwidth/2-viadim_osy)
		viayofs = to_float(viayofs if vdim_center > 0 else -1*viayofs)
		viayofs = viayofs if viaoffset[1] else 0
		h_to_v_via_ref.movex(viaxofs).movey(viayofs)
	# add ports and return
	Lroute.add_ports(h_to_v_via_ref.get_ports_list())
	return rename_ports_by_orientation(Lroute.flatten())





from glayout.pdk.mappedpdk import MappedPDK
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.rectangle import rectangle
from gdsfactory.components.rectangular_ring import rectangular_ring
from glayout.primitives.via_gen import via_array, via_stack
from typing import Optional
from glayout.pdk.util.comp_utils import to_decimal, to_float, evaluate_bbox
from glayout.pdk.util.port_utils import print_ports
from glayout.pdk.util.snap_to_grid import component_snap_to_grid
from glayout.routing.L_route import L_route


@cell
def tapring(
    pdk: MappedPDK,
    enclosed_rectangle=(2.0, 4.0),
    sdlayer: str = "p+s/d",
    horizontal_glayer: str = "met2",
    vertical_glayer: str = "met1",
    sides: tuple[bool,bool,bool,bool] = (True,True,True,True)
) -> Component:
    """ptapring produce a p substrate / pwell tap rectanglular ring
    This ring will legally enclose a rectangular shape
    args:
    pdk: MappedPDK is the pdk to use
    enclosed_rectangle: tuple is the (width, hieght) of the area to enclose
    ****NOTE: the enclosed_rectangle will be the enclosed dimensions of active_tap
    horizontal_glayer: string=met2, layer used over the ring horizontally
    vertical_glayer: string=met1, layer used over the ring vertically
    sides: instead of creating the ring on all sides, only create it on the specified sides (W,N,E,S)
    ports:
    Narr_... all ports in top via array
    Sarr_... all ports in bottom via array
    Earr_... all ports in right via array
    Warr_... all ports in left via array
    bl_corner_...all ports in bottom left L route
    """
    enclosed_rectangle = pdk.snap_to_2xgrid(enclosed_rectangle,return_type="float")
    # check layers, activate pdk, create top cell
    pdk.has_required_glayers(
        [sdlayer, "active_tap", "mcon", horizontal_glayer, vertical_glayer]
    )
    pdk.activate()
    ptapring = Component()
    if not "met" in horizontal_glayer or not "met" in vertical_glayer:
        raise ValueError("both horizontal and vertical glayers should be metals")
    # check that ring is not too small
    min_gap_tap = pdk.get_grule("active_tap")["min_separation"]
    if enclosed_rectangle[0] < min_gap_tap:
        raise ValueError("ptapring must be larger than " + str(min_gap_tap))
    # create active tap
    tap_width = max(
        pdk.get_grule("active_tap")["min_width"],
        2 * pdk.get_grule("active_tap", "mcon")["min_enclosure"]
        + pdk.get_grule("mcon")["width"],
    )
    ptapring << rectangular_ring(
        enclosed_size=enclosed_rectangle,
        width=tap_width,
        centered=True,
        layer=pdk.get_glayer("active_tap"),
    )
    # create p plus area
    pp_enclosure = pdk.get_grule("active_tap", sdlayer)["min_enclosure"]
    pp_width = 2 * pp_enclosure + tap_width
    pp_enclosed_rectangle = [dim - 2 * pp_enclosure for dim in enclosed_rectangle]
    ptapring << rectangular_ring(
        enclosed_size=pp_enclosed_rectangle,
        width=pp_width,
        centered=True,
        layer=pdk.get_glayer(sdlayer),
    )
    # create via arrs
    via_width_horizontal = evaluate_bbox(via_stack(pdk, "active_tap", horizontal_glayer))[0]
    arr_size_horizontal = enclosed_rectangle[0]
    horizontal_arr = via_array(
        pdk,
        "active_tap",
        horizontal_glayer,
        (arr_size_horizontal, via_width_horizontal),
        minus1=True,
        lay_every_layer=True
    )
    via_width_vertical = evaluate_bbox(via_stack(pdk, "active_tap", vertical_glayer))[1]
    arr_size_vertical = enclosed_rectangle[1]
    vertical_arr = via_array(
        pdk,
        "active_tap",
        vertical_glayer,
        (via_width_vertical, arr_size_vertical),
        minus1=True,
        lay_every_layer=True
    )
    # add via arrs
    refs_prefixes = list()
    if sides[1]:
        metal_ref_n = ptapring << horizontal_arr
        metal_ref_n.movey(round(0.5 * (enclosed_rectangle[1] + tap_width),4))
        refs_prefixes.append((metal_ref_n,"N_"))
    if sides[2]:
        metal_ref_e = ptapring << vertical_arr
        metal_ref_e.movex(round(0.5 * (enclosed_rectangle[0] + tap_width),4))
        refs_prefixes.append((metal_ref_e,"E_"))
    if sides[3]:
        metal_ref_s = ptapring << horizontal_arr
        metal_ref_s.movey(round(-0.5 * (enclosed_rectangle[1] + tap_width),4))
        refs_prefixes.append((metal_ref_s,"S_"))
    if sides[0]:
        metal_ref_w = ptapring << vertical_arr
        metal_ref_w.movex(round(-0.5 * (enclosed_rectangle[0] + tap_width),4))
        refs_prefixes.append((metal_ref_w,"W_"))
    # connect vertices
    if sides[1] and sides[0]:
        tlvia = ptapring << L_route(pdk, metal_ref_n.ports["top_met_W"], metal_ref_w.ports["top_met_N"])
        refs_prefixes += [(tlvia,"tl_")]
    if sides[1] and sides[2]:
        trvia = ptapring << L_route(pdk, metal_ref_n.ports["top_met_E"], metal_ref_e.ports["top_met_N"])
        refs_prefixes += [(trvia,"tr_")]
    if sides[3] and sides[0]:
        blvia = ptapring << L_route(pdk, metal_ref_s.ports["top_met_W"], metal_ref_w.ports["top_met_S"])
        refs_prefixes += [(blvia,"bl_")]
    if sides[3] and sides[2]:
        brvia = ptapring << L_route(pdk, metal_ref_s.ports["top_met_E"], metal_ref_e.ports["top_met_S"])
        refs_prefixes += [(brvia,"br_")]
    # add ports, flatten and return
    for ref_, prefix in refs_prefixes:
        ptapring.add_ports(ref_.get_ports_list(),prefix=prefix)
    return component_snap_to_grid(ptapring)






from gdsfactory.cell import cell, clear_cache
from gdsfactory.component import Component, copy
from gdsfactory.component_reference import ComponentReference
from gdsfactory.components.rectangle import rectangle
from glayout.pdk.mappedpdk import MappedPDK
from typing import Optional, Union
from glayout.primitives.fet import nmos, pmos, multiplier
from glayout.components.diff_pair import diff_pair
from glayout.primitives.guardring import tapring
from glayout.primitives.mimcap import mimcap_array, mimcap
from glayout.routing.L_route import L_route
from glayout.routing.c_route import c_route
from glayout.primitives.via_gen import via_stack, via_array
from gdsfactory.routing.route_quad import route_quad
from glayout.pdk.util.comp_utils import (
    evaluate_bbox,
    prec_ref_center,
    movex,
    movey,
    to_decimal,
    to_float,
    move,
    align_comp_to_port,
    get_padding_points_cc,
)
from glayout.pdk.util.port_utils import (
    rename_ports_by_orientation,
    rename_ports_by_list,
    add_ports_perimeter,
    print_ports,
    set_port_orientation,
    rename_component_ports,
)
from glayout.routing.straight_route import straight_route
from glayout.pdk.util.snap_to_grid import component_snap_to_grid
from pydantic import validate_arguments
from glayout.placement.two_transistor_interdigitized import two_nfet_interdigitized
from glayout.spice import Netlist
from glayout.components.stacked_current_mirror import current_mirror_netlist

def diff_pair_ibias_netlist(center_diffpair: Component, current_mirror: Component, antenna_diode: Component) -> Netlist:
    netlist = Netlist(
        circuit_name="DIFFPAIR_CMIRROR_BIAS",
        nodes=['VP', 'VN', 'VDD1', 'VDD2', 'IBIAS', 'VSS', 'B']
    )

    diffpair_ref = netlist.connect_netlist(
        center_diffpair.info['netlist'],
        []
    )

    cmirror_ref = netlist.connect_netlist(
        current_mirror.info['netlist'],
        [('VREF', 'IBIAS')]
    )

    netlist.connect_subnets(
        cmirror_ref,
        diffpair_ref,
        [('VCOPY', 'VTAIL')]
    )

    netlist.connect_netlist(
        antenna_diode.info['netlist'],
        [('D', 'VSS'), ('G', 'VSS'), ('B', 'VSS'), ('S', 'VP')]
    )

    netlist.connect_netlist(
        antenna_diode.info['netlist'],
        [('D', 'VSS'), ('G', 'VSS'), ('B', 'VSS'), ('S', 'VN')]
    )

    return netlist

@validate_arguments
def diff_pair_ibias(
    pdk: MappedPDK,
    half_diffpair_params: tuple[float, float, int],
    diffpair_bias: tuple[float, float, int],
    rmult: int,
    with_antenna_diode_on_diffinputs: int,
) -> Component:
    # create and center diffpair
    diffpair_i_ = Component("temp diffpair and current source")
    center_diffpair_comp = diff_pair(
        pdk,
        width=half_diffpair_params[0],
        length=half_diffpair_params[1],
        fingers=half_diffpair_params[2],
        rmult=rmult,
    )
    # add antenna diodes if that option was specified
    diffpair_centered_ref = prec_ref_center(center_diffpair_comp)
    diffpair_i_.add(diffpair_centered_ref)
    diffpair_i_.add_ports(diffpair_centered_ref.get_ports_list())
    if with_antenna_diode_on_diffinputs:
        antenna_diode_comp = nmos(
            pdk,
            1,
            with_antenna_diode_on_diffinputs,
            1,
            with_dummy=False,
            with_tie=False,
            with_substrate_tap=False,
            with_dnwell=False,
            length=0.5,
            sd_route_topmet="met2",
            gate_route_topmet="met1",
        ).copy()
        antenna_diode_comp << straight_route(
            pdk,
            antenna_diode_comp.ports["multiplier_0_row0_col0_rightsd_top_met_S"],
            antenna_diode_comp.ports["multiplier_0_gate_N"],
        )
        antenna_diode_refL = diffpair_i_ << antenna_diode_comp
        antenna_diode_refR = diffpair_i_ << antenna_diode_comp
        align_comp_to_port(
            antenna_diode_refL, diffpair_i_.ports["MINUSgateroute_W_con_N"], ("r", "t")
        )
        antenna_diode_refL.movex(pdk.util_max_metal_seperation())
        align_comp_to_port(
            antenna_diode_refR, diffpair_i_.ports["MINUSgateroute_E_con_N"], ("L", "t")
        )
        antenna_diode_refR.movex(0 - pdk.util_max_metal_seperation())
        # route the antenna diodes to gnd and
        Lgndcon = diffpair_i_.ports["tap_W_top_met_N"]
        Lgndcon.layer = pdk.get_glayer("met1")
        Rgndcon = diffpair_i_.ports["tap_E_top_met_N"]
        Rgndcon.layer = pdk.get_glayer("met1")
        diffpair_i_ << L_route(
            pdk, antenna_diode_refL.ports["multiplier_0_gate_E"], Lgndcon
        )
        diffpair_i_ << L_route(
            pdk, antenna_diode_refR.ports["multiplier_0_gate_W"], Rgndcon
        )
        diffpair_i_ << straight_route(
            pdk,
            antenna_diode_refL.ports["multiplier_0_source_W"],
            diffpair_i_.ports["MINUSgateroute_W_con_N"],
        )
        diffpair_i_ << straight_route(
            pdk,
            antenna_diode_refR.ports["multiplier_0_source_W"],
            diffpair_i_.ports["PLUSgateroute_E_con_N"],
        )
    # create and position tail current source
    cmirror = two_nfet_interdigitized(
        pdk,
        width=diffpair_bias[0],
        length=diffpair_bias[1],
        numcols=diffpair_bias[2],
        with_tie=True,
        with_substrate_tap=False,
        gate_route_topmet="met3",
        sd_route_topmet="met3",
        rmult=rmult,
        tie_layers=("met2", "met2"),
    )
    # cmirror routing
    metal_sep = pdk.util_max_metal_seperation()
    gate_short = cmirror << c_route(
        pdk,
        cmirror.ports["A_gate_E"],
        cmirror.ports["B_gate_E"],
        extension=3 * metal_sep,
        viaoffset=None,
    )
    cmirror << L_route(
        pdk,
        gate_short.ports["con_N"],
        cmirror.ports["A_drain_E"],
        viaoffset=False,
        fullbottom=False,
    )
    srcshort = cmirror << c_route(
        pdk,
        cmirror.ports["A_source_W"],
        cmirror.ports["B_source_W"],
        extension=metal_sep,
        viaoffset=False,
    )
    cmirror.add_ports(srcshort.get_ports_list(), prefix="purposegndports")
    # current mirror netlist
    cmirror.info['netlist'] = current_mirror_netlist(
        pdk,
        width=diffpair_bias[0],
        length=diffpair_bias[1],
        multipliers=diffpair_bias[2]
    )

    # add cmirror
    tailcurrent_ref = diffpair_i_ << cmirror
    tailcurrent_ref.movey(
        pdk.snap_to_2xgrid(
            -0.5 * (center_diffpair_comp.ymax - center_diffpair_comp.ymin)
            - abs(tailcurrent_ref.ymax)
            - metal_sep
        )
    )
    purposegndPort = tailcurrent_ref.ports["purposegndportscon_S"].copy()
    purposegndPort.name = "ibias_purposegndport"
    diffpair_i_.add_ports([purposegndPort])
    diffpair_i_.add_ports(tailcurrent_ref.get_ports_list(), prefix="ibias_")

    diffpair_i_ref = prec_ref_center(diffpair_i_)

    diffpair_i_ref.info['netlist'] = diff_pair_ibias_netlist(center_diffpair_comp, cmirror, antenna_diode_comp)
    return diffpair_i_ref




from gdsfactory.cell import cell, clear_cache
from gdsfactory.component import Component, copy
from gdsfactory.component_reference import ComponentReference
from gdsfactory.components.rectangle import rectangle
from glayout.pdk.mappedpdk import MappedPDK
from typing import Optional, Union
from glayout.primitives.fet import nmos, pmos, multiplier
from glayout.components.diff_pair import diff_pair
from glayout.primitives.guardring import tapring
from glayout.primitives.mimcap import mimcap_array, mimcap
from glayout.routing.L_route import L_route
from glayout.routing.c_route import c_route
from glayout.primitives.via_gen import via_stack, via_array
from gdsfactory.routing.route_quad import route_quad
from glayout.pdk.util.comp_utils import evaluate_bbox, prec_ref_center, movex, movey, to_decimal, to_float, move, align_comp_to_port, get_padding_points_cc
from glayout.pdk.util.port_utils import rename_ports_by_orientation, rename_ports_by_list, add_ports_perimeter, print_ports, set_port_orientation, rename_component_ports
from glayout.routing.straight_route import straight_route
from glayout.pdk.util.snap_to_grid import component_snap_to_grid
from pydantic import validate_arguments
from glayout.placement.two_transistor_interdigitized import two_nfet_interdigitized
from glayout.spice import Netlist

def row_csamplifier_diff_to_single_ended_converter_netlist(diff_to_single: Component) -> Netlist:
    overall_netlist = Netlist(
        circuit_name="DIFF_TO_SINGLE_CS",
        nodes=['VIN1', 'VIN2', 'VOUT', 'VSS', 'VSS2']
    )

    overall_netlist.connect_netlist(
        diff_to_single.info['netlist'],
        [('VIN', 'VIN1'), ('VOUT', 'VIN2')]
    )

    return overall_netlist

def __connect_cs_netlist(pmos_comps: Component, half_cs_pmos: Component):
    pmos_comps.info['netlist'].connect_netlist(
        half_cs_pmos.info['netlist'],
        [('D', 'VOUT'), ('S', 'VSS'), ('B', 'VSS'), ('G', 'VIN2')]
    )

def row_csamplifier_diff_to_single_ended_converter(pdk: MappedPDK, diff_to_single_ended_converter: Component, pamp_hparams, rmult) -> Component:
    pmos_comps = diff_to_single_ended_converter

    pmos_comps.info['netlist'] = row_csamplifier_diff_to_single_ended_converter_netlist(diff_to_single_ended_converter)

    x_dim_center = max(abs(pmos_comps.xmax),abs(pmos_comps.xmin))
    for direction in [-1, 1]:
        halfMultp = pmos(
            pdk,
            width=pamp_hparams[0],
            length=pamp_hparams[1],
            fingers=pamp_hparams[2],
            multipliers=pamp_hparams[3],
            with_tie=True,
            dnwell=False,
            with_substrate_tap=False,
            sd_route_left=bool(direction-1),
            rmult=rmult,
            tie_layers=("met2","met2")
        )
        halfMultp_ref = pmos_comps << halfMultp
        halfMultp_ref.movex(direction * abs(x_dim_center + halfMultp_ref.xmax+1))
        label = "L_" if direction==-1 else "R_"
        # this special marker is used to rename these ports in the opamp to commonsource_Pamp_
        pmos_comps.add_ports(halfMultp_ref.get_ports_list(),prefix="halfpspecialmarker_"+label)

        __connect_cs_netlist(pmos_comps, halfMultp)

    # add npadding and add ports
    nwellbbox = pmos_comps.extract(layers=[pdk.get_glayer("poly"),pdk.get_glayer("active_diff"),pdk.get_glayer("active_tap"), pdk.get_glayer("nwell"),pdk.get_glayer("dnwell")]).bbox
    nwellspacing = pdk.get_grule("nwell", "active_tap")["min_enclosure"]
    nwell_points = get_padding_points_cc(nwellbbox, default=nwellspacing, pdk_for_snap2xgrid=pdk)
    pmos_comps.add_polygon(nwell_points, layer=pdk.get_glayer("nwell"))
    tapcenter_rect = [(evaluate_bbox(pmos_comps)[0] + 1), (evaluate_bbox(pmos_comps)[1] + 1)]
    topptap = prec_ref_center(tapring(pdk, tapcenter_rect, "p+s/d",vertical_glayer="met2"),destination=tuple(pmos_comps.center))
    pmos_comps.add(topptap)
    pmos_comps.add_ports(topptap.get_ports_list(),prefix="top_ptap_")
    # vdd taprings of the center components
    pmos_comps << straight_route(pdk, pmos_comps.ports["ptopAB_L_welltap_W_top_met_W"],pmos_comps.ports["halfpspecialmarker_L_tie_E_top_met_N"],width=2,glayer1="met2",via1_alignment=('c','c'),fullbottom=True)
    pmos_comps << straight_route(pdk, pmos_comps.ports["ptopAB_R_welltap_E_top_met_E"],pmos_comps.ports["halfpspecialmarker_R_tie_W_top_met_N"],width=2,glayer1="met2",via1_alignment=('c','c'),fullbottom=True)
    pmos_comps << straight_route(pdk, pmos_comps.ports["pbottomAB_L_welltap_W_top_met_W"],pmos_comps.ports["halfpspecialmarker_L_tie_E_top_met_W"],width=2,glayer1="met2",via1_alignment=('c','c'),via2_alignment=('c','c'),fullbottom=True)
    pmos_comps << straight_route(pdk, pmos_comps.ports["pbottomAB_R_welltap_E_top_met_E"],pmos_comps.ports["halfpspecialmarker_R_tie_W_top_met_E"],width=2,glayer1="met2",via1_alignment=('c','c'),via2_alignment=('c','c'),fullbottom=True)
    return pmos_comps




from gdsfactory.read.import_gds import import_gds
from gdsfactory.component import Component
from pathlib import Path
import os
import math
from gdsfactory.pdk import Pdk
from pathlib import Path
from typing import Union, Optional
from pydantic import validate_arguments

def get_files_with_extension(directory, extension):
	file_list = []
	for filename in os.listdir(directory):
		if filename.endswith(extension):
			file_list.append(filename)
	return file_list


@validate_arguments
def write_component_matrix(components_dir: Union[str,Path,list]="./", xspace: float=400,yspace: float=280, rtr_comp: bool=False, write_name: str="big_gds_here.gds"):
	"""Use the write_component_matrix function to create a matrix of many different components
	reads the different components from all gds files in components_dir
	args:
	components_dir = a file directory where all gds files are treated as components (i.e. to add to the matrix)
	****Note: you can specify this as a list Components, in which case, the list is used to make the matrix
	xspace = xspacing to use (center to center x distance between adajacent elements in the matrix)
	yspace = yspacing to use (center to center y distance between adajacent elements in the matrix)
	rtr_comp = if true will not write the component to gds (default = false)
	write_name = name/path of gds write file
	"""
	pdk_nochache = Pdk(name="nocache")
	pdk_nochache.cell_decorator_settings.cache=False
	pdk_nochache.activate()

	if isinstance(components_dir, list):
		c_comp_list = components_dir
	else:
		search_dir = Path(components_dir).resolve()
		c_files_list = get_files_with_extension(str(search_dir),".gds")
		c_comp_list = list()
		for i,filev in enumerate(c_files_list):
			if "big_gds_here" in str(filev) or write_name in str(filev):
				continue
			tempcomp = import_gds(filev)
			tempcomp.name = "circ"+str(i)
			c_comp_list.append(tempcomp)

	col_len = round(math.sqrt(len(c_comp_list)))
	col_index = 0
	row_index = 0
	big_comp = Component("big comp")
	for comp_v in c_comp_list:
		if comp_v is None:
			continue
		opref = big_comp << comp_v
		opref.movex(col_index * xspace).movey(row_index*yspace)
		col_index += 1
		if not col_index % col_len:
			col_index=0
			row_index += 1
	if rtr_comp:
		return big_comp
	else:
		big_comp.write_gds(write_name)

if __name__=="__main__":
	write_component_matrix()
